"""
_RunConfigAPI_

API for anyting RunConfig related

"""
import logging
import threading
import os.path
import time
from datetime import datetime

from Utils.Utilities import rootUrlJoin

from WMCore.DAOFactory import DAOFactory

from WMCore.WorkQueue.WMBSHelper import WMBSHelper
from WMCore.WMBS.Fileset import Fileset

from WMCore.ReqMgr.DataStructs.RequestStatus import REQUEST_START_STATE

from T0.RunConfig.Tier0Config import retrieveDatasetConfig
from T0.RunConfig.Tier0Config import retrieveSiteConfig
from T0.RunConfig.Tier0Config import addRepackConfig
from T0.RunConfig.Tier0Config import deleteStreamConfig

from WMCore.WMSpec.StdSpecs.Repack import RepackWorkloadFactory
from WMCore.WMSpec.StdSpecs.Express import ExpressWorkloadFactory
from WMCore.WMSpec.StdSpecs.PromptReco import PromptRecoWorkloadFactory

def extractConfigParameter(configParameter, era, run):
    """
    _extractConfigParameter_

    Checks if configParameter is era or run dependent. If it is, use the
    provided era and run information to extract the correct parameter.
    """
    if isinstance(configParameter, dict):
        if 'acqEra' in configParameter and era in configParameter['acqEra']:
            return configParameter['acqEra'][era]
        elif 'maxRun' in configParameter:
            newConfigParameter = None
            for maxRun in sorted(configParameter['maxRun'].keys()):
                if run <= maxRun:
                    newConfigParameter = configParameter['maxRun'][maxRun]
                    break
            if newConfigParameter:
                return newConfigParameter
        return configParameter['default']
    else:
        return configParameter

def configureRun(tier0Config, run, hltConfig, referenceHltConfig = None):
    """
    _configureRun_

    Called by Tier0Feeder for new runs.

    Retrieve HLT config and configure global run
    settings and stream/dataset/trigger mapping

    """
    logging.debug("configureRun() : %d" % run)
    myThread = threading.currentThread()

    daoFactory = DAOFactory(package = "T0.WMBS",
                            logger = logging,
                            dbinterface = myThread.dbi)

    # dao to update global run settings
    updateRunDAO = daoFactory(classname = "RunConfig.UpdateRun")

    # workaround to make unit test work without HLTConfDatabase
    if hltConfig == None and referenceHltConfig != None:
        hltConfig = referenceHltConfig

    # treat centralDAQ or miniDAQ runs (have an HLT key) different from local runs
    if hltConfig != None:

        # write stream/dataset/trigger mapping
        insertStreamDAO = daoFactory(classname = "RunConfig.InsertStream")
        insertDatasetDAO = daoFactory(classname = "RunConfig.InsertPrimaryDataset")
        insertStreamDatasetDAO = daoFactory(classname = "RunConfig.InsertStreamDataset")
        insertTriggerDAO = daoFactory(classname = "RunConfig.InsertTrigger")
        insertDatasetTriggerDAO = daoFactory(classname = "RunConfig.InsertDatasetTrigger")

        # partition AlcaHarvest upload by year
        if tier0Config.Global.AlcaHarvestCondLFNBase:
            alcaHarvestCondLFNBase = os.path.join(tier0Config.Global.AlcaHarvestCondLFNBase, str(datetime.now().year))
        if tier0Config.Global.AlcaHarvestLumiURL:
            alcaHarvestLumiURL = rootUrlJoin(tier0Config.Global.AlcaHarvestLumiURL, str(datetime.now().year))
            if not alcaHarvestLumiURL:
                raise RuntimeError("Problem in configureRun() : Invalid AlcaHarvestLumiURL !")

        bindsUpdateRun = { 'RUN' : run,
                           'PROCESS' : hltConfig['process'],
                           'ACQERA' : tier0Config.Global.AcquisitionEra,
                           'BACKFILL' : tier0Config.Global.Backfill,
                           'BULKDATATYPE' : tier0Config.Global.BulkDataType,
                           'DQMUPLOADURL' : tier0Config.Global.DQMUploadUrl,
                           'AHTIMEOUT' : tier0Config.Global.AlcaHarvestTimeout,
                           'AHCONDLFNBASE' : alcaHarvestCondLFNBase,
                           'AHLUMIURL' : alcaHarvestLumiURL,
                           'CONDTIMEOUT' : tier0Config.Global.ConditionUploadTimeout,
                           'DBHOST' : tier0Config.Global.DropboxHost,
                           'VALIDMODE' : tier0Config.Global.ValidationMode }

        bindsStream = []
        bindsDataset = []
        bindsStreamDataset = []
        bindsTrigger = []
        bindsDatasetTrigger = []

        for stream, datasetDict in list(hltConfig['mapping'].items()):
            bindsStream.append( { 'STREAM' : stream } )
            for dataset, paths in list(datasetDict.items()):

                if dataset == "Unassigned path":

                    if run < 317512:
                        pass
                    else:
                        raise RuntimeError("Problem in configureRun() : Unassigned path in HLT menu !")

                else:
                    bindsDataset.append( { 'PRIMDS' : dataset } )
                    bindsStreamDataset.append( { 'RUN' : run,
                                                 'PRIMDS' : dataset,
                                                 'STREAM' : stream } )
                    for path in paths:
                        bindsTrigger.append( { 'TRIG' : path } )
                        bindsDatasetTrigger.append( { 'RUN' : run,
                                                      'TRIG' : path,
                                                      'PRIMDS' : dataset } )

        try:
            myThread.transaction.begin()
            updateRunDAO.execute(bindsUpdateRun, conn = myThread.transaction.conn, transaction = True)
            insertStreamDAO.execute(bindsStream, conn = myThread.transaction.conn, transaction = True)
            insertDatasetDAO.execute(bindsDataset, conn = myThread.transaction.conn, transaction = True)
            insertStreamDatasetDAO.execute(bindsStreamDataset, conn = myThread.transaction.conn, transaction = True)
            insertTriggerDAO.execute(bindsTrigger, conn = myThread.transaction.conn, transaction = True)
            insertDatasetTriggerDAO.execute(bindsDatasetTrigger, conn = myThread.transaction.conn, transaction = True)
        except Exception as ex:
            logging.exception(ex)
            myThread.transaction.rollback()
            raise RuntimeError("Problem in configureRun() database transaction !")
        else:
            myThread.transaction.commit()

    else:

        bindsUpdateRun = { 'RUN' : run,
                           'PROCESS' : "FakeProcessName",
                           'ACQERA' : "FakeAcquisitionEra",
                           'BACKFILL' : None,
                           'BULKDATATYPE' : "FakeBulkDataType",
                           'AHTIMEOUT' : None,
                           'AHDIR' : None,
                           'CONDTIMEOUT' : None,
                           'DBHOST' : None,
                           'VALIDMODE' : None }

        try:
            myThread.transaction.begin()
            updateRunDAO.execute(bindsUpdateRun, conn = myThread.transaction.conn, transaction = True)
        except Exception as ex:
            logging.exception(ex)
            myThread.transaction.rollback()
            raise RuntimeError("Problem in configureRun() database transaction !")
        else:
            myThread.transaction.commit()

    return

def configureRunStream(tier0Config, run, stream, specDirectory, dqmUploadProxy):
    """
    _configureRunStream_

    Called by Tier0Feeder for new run/streams.

    Retrieve global run settings and build the part
    of the configuration relevant to run/stream
    and write it to the database.

    Create workflows, filesets and subscriptions for
    the processing of runs/streams.

    """
    logging.debug("configureRunStream() : %d , %s" % (run, stream))
    myThread = threading.currentThread()

    daoFactory = DAOFactory(package = "T0.WMBS",
                            logger = logging,
                            dbinterface = myThread.dbi)

    # retrieve some basic run information
    getRunInfoDAO = daoFactory(classname = "RunConfig.GetRunInfo")
    runInfo = getRunInfoDAO.execute(run, transaction = False)[0]

    # treat centralDAQ or miniDAQ runs (have an HLT key) different from local runs
    if runInfo['hltkey'] != None:

        # streams not explicitely configured are repacked
        if stream not in list(tier0Config.Streams.dictionary_().keys()):
            addRepackConfig(tier0Config, stream)

        streamConfig = tier0Config.Streams.dictionary_()[stream]

        # consistency check to make sure stream exists and has datasets defined
        # only run if we don't ignore the stream
        if streamConfig.ProcessingStyle != "Ignore":
            getStreamDatasetsDAO = daoFactory(classname = "RunConfig.GetStreamDatasets")
            datasets = getStreamDatasetsDAO.execute(run, stream, transaction = False)
            if len(datasets) == 0:
                raise RuntimeError("Stream is not defined in HLT menu or has no datasets !")

        # write run/stream processing completion record
        insertRunStreamDoneDAO = daoFactory(classname = "RunConfig.InsertRunStreamDone")

        # write stream/dataset mapping (for special express and error datasets)
        insertDatasetDAO = daoFactory(classname = "RunConfig.InsertPrimaryDataset")
        insertStreamDatasetDAO = daoFactory(classname = "RunConfig.InsertStreamDataset")

        # write stream configuration
        insertCMSSWVersionDAO = daoFactory(classname = "RunConfig.InsertCMSSWVersion")
        insertStreamStyleDAO = daoFactory(classname = "RunConfig.InsertStreamStyle")
        insertRepackConfigDAO = daoFactory(classname = "RunConfig.InsertRepackConfig")
        insertPromptCalibrationDAO = daoFactory(classname = "RunConfig.InsertPromptCalibration")
        insertExpressConfigDAO = daoFactory(classname = "RunConfig.InsertExpressConfig")
        insertSpecialDatasetDAO = daoFactory(classname = "RunConfig.InsertSpecialDataset")
        insertDatasetScenarioDAO = daoFactory(classname = "RunConfig.InsertDatasetScenario")
        insertStreamFilesetDAO = daoFactory(classname = "RunConfig.InsertStreamFileset")
        insertRecoReleaseConfigDAO = daoFactory(classname = "RunConfig.InsertRecoReleaseConfig")
        insertWorkflowMonitoringDAO = daoFactory(classname = "RunConfig.InsertWorkflowMonitoring")
        insertStorageNodeDAO = daoFactory(classname = "RunConfig.InsertStorageNode")
        insertPhEDExConfigDAO = daoFactory(classname = "RunConfig.InsertPhEDExConfig")

        bindsRunStreamDone = {'RUN' : run,
                              'STREAM' : stream}
        bindsCMSSWVersion = []
        bindsDataset = []
        bindsStreamDataset = []
        bindsStreamStyle = {'RUN' : run,
                            'STREAM' : stream,
                            'STYLE': streamConfig.ProcessingStyle }
        bindsRepackConfig = {}
        bindsPromptCalibration = {}
        bindsExpressConfig = {}
        bindsSpecialDataset = {}
        bindsDatasetScenario = []
        bindsStorageNode = []
        bindsPhEDExConfig = []

        #
        # for spec creation, details for all outputs
        #
        outputModuleDetails = []

        #
        # special dataset for some express output
        #
        specialDataset = None

        #
        # for PhEDEx subscription settings
        #
        subscriptions = []

        #
        # first take care of all stream settings
        #
        getStreamOnlineVersionDAO = daoFactory(classname = "RunConfig.GetStreamOnlineVersion")
        onlineVersion = getStreamOnlineVersionDAO.execute(run, stream, transaction = False)

        if streamConfig.ProcessingStyle == "Bulk":

            streamConfig.Repack.CMSSWVersion = streamConfig.VersionOverride.get(onlineVersion, onlineVersion)

            bindsCMSSWVersion.append( { 'VERSION' : streamConfig.Repack.CMSSWVersion } )

            streamConfig.Repack.ScramArch = tier0Config.Global.ScramArches.get(streamConfig.Repack.CMSSWVersion,
                                                                               tier0Config.Global.DefaultScramArch)

            # check for era or run dependent config parameters
            streamConfig.Repack.ProcessingVersion = extractConfigParameter(streamConfig.Repack.ProcessingVersion, runInfo['acq_era'], run)

            bindsRepackConfig = { 'RUN' : run,
                                  'STREAM' : stream,
                                  'PROC_VER': streamConfig.Repack.ProcessingVersion,
                                  'MAX_SIZE_SINGLE_LUMI' : streamConfig.Repack.MaxSizeSingleLumi,
                                  'MAX_SIZE_MULTI_LUMI' : streamConfig.Repack.MaxSizeMultiLumi,
                                  'MIN_SIZE' : streamConfig.Repack.MinInputSize,
                                  'MAX_SIZE' : streamConfig.Repack.MaxInputSize,
                                  'MAX_EDM_SIZE' : streamConfig.Repack.MaxEdmSize,
                                  'MAX_OVER_SIZE' : streamConfig.Repack.MaxOverSize,
                                  'MAX_EVENTS' : streamConfig.Repack.MaxInputEvents,
                                  'MAX_FILES' : streamConfig.Repack.MaxInputFiles,
                                  'CMSSW' : streamConfig.Repack.CMSSWVersion,
                                  'SCRAM_ARCH' : streamConfig.Repack.ScramArch }

        elif streamConfig.ProcessingStyle == "Express":

            # check for era or run dependent config parameters
            streamConfig.Express.Scenario = extractConfigParameter(streamConfig.Express.Scenario, runInfo['acq_era'], run)

            specialDataset = "Stream%s" % stream
            bindsDataset.append( { 'PRIMDS' : specialDataset } )
            bindsStreamDataset.append( { 'RUN' : run,
                                         'PRIMDS' : specialDataset,
                                         'STREAM' : stream } )
            bindsSpecialDataset = { 'STREAM' : stream,
                                    'PRIMDS' : specialDataset }
            bindsDatasetScenario.append( { 'RUN' : run,
                                           'PRIMDS' : specialDataset,
                                           'SCENARIO' : streamConfig.Express.Scenario } )

            if streamConfig.Express.WriteDQM:
                outputModuleDetails.append( { 'dataTier' : tier0Config.Global.DQMDataTier,
                                              'eventContent' : tier0Config.Global.DQMDataTier,
                                              'primaryDataset' : specialDataset } )

            if streamConfig.Express.ArchivalNode or streamConfig.Express.TapeNode or streamConfig.Express.DiskNode:

                bindsPhEDExConfig.append( { 'RUN' : run,
                                            'PRIMDS' : specialDataset,
                                            'ARCHIVAL_NODE' : streamConfig.Express.ArchivalNode,
                                            'TAPE_NODE' : streamConfig.Express.TapeNode,
                                            'DISK_NODE' :  streamConfig.Express.DiskNode,
                                            'DISK_NODE_RECO' : None } )

                custodialSites = []
                nonCustodialSites = []
                autoApproveSites = []
                if streamConfig.Express.ArchivalNode:
                    bindsStorageNode.append( { 'NODE' : streamConfig.Express.ArchivalNode } )
                    custodialSites.append(streamConfig.Express.ArchivalNode)
                    autoApproveSites.append(streamConfig.Express.ArchivalNode)
                if streamConfig.Express.TapeNode:
                    bindsStorageNode.append( { 'NODE' : streamConfig.Express.TapeNode } )
                    custodialSites.append(streamConfig.Express.TapeNode)
                if streamConfig.Express.DiskNode:
                    bindsStorageNode.append( { 'NODE' : streamConfig.Express.DiskNode } )
                    nonCustodialSites.append(streamConfig.Express.DiskNode)
                    autoApproveSites.append(streamConfig.Express.DiskNode)

                if len(custodialSites) > 0 or len(nonCustodialSites) > 0:
                    subscriptions.append( { 'custodialSites' : custodialSites,
                                            'custodialSubType' : "Replica",
                                            'custodialGroup' : "DataOps",
                                            'nonCustodialSites' : nonCustodialSites,
                                            'nonCustodialSubType' : "Replica",
                                            'nonCustodialGroup' : streamConfig.Express.PhEDExGroup,
                                            'autoApproveSites' : autoApproveSites,
                                            'priority' : "high",
                                            'primaryDataset' : specialDataset,
                                            'deleteFromSource' : True } )

            alcaSkim = None
            if len(streamConfig.Express.AlcaSkims) > 0:
                outputModuleDetails.append( { 'dataTier' : "ALCARECO",
                                              'eventContent' : "ALCARECO",
                                              'primaryDataset' : specialDataset } )
                alcaSkim = ",".join(streamConfig.Express.AlcaSkims)

                numPromptCalibProd = 0
                for producer in streamConfig.Express.AlcaSkims:
                    if producer.startswith("PromptCalibProd"):
                        numPromptCalibProd += 1

                if numPromptCalibProd > 0:
                    bindsPromptCalibration = { 'RUN' : run,
                                               'STREAM' : stream,
                                               'NUM_PRODUCER' : numPromptCalibProd }

            dqmSeq = None
            if len(streamConfig.Express.DqmSequences) > 0:
                dqmSeq = ",".join(streamConfig.Express.DqmSequences)

            streamConfig.Express.CMSSWVersion = streamConfig.VersionOverride.get(onlineVersion, onlineVersion)

            bindsCMSSWVersion.append( { 'VERSION' : streamConfig.Express.CMSSWVersion } )

            streamConfig.Express.ScramArch = tier0Config.Global.ScramArches.get(streamConfig.Express.CMSSWVersion,
                                                                                tier0Config.Global.DefaultScramArch)
            
            streamConfig.Express.RecoScramArch = None
            if streamConfig.Express.RecoCMSSWVersion != None:

                # check for era or run dependent config parameters
                streamConfig.Express.RecoCMSSWVersion = extractConfigParameter(streamConfig.Express.RecoCMSSWVersion, runInfo['acq_era'], run)

                bindsCMSSWVersion.append( { 'VERSION' : streamConfig.Express.RecoCMSSWVersion } )

                streamConfig.Express.RecoScramArch = tier0Config.Global.ScramArches.get(streamConfig.Express.RecoCMSSWVersion,
                                                                                        tier0Config.Global.DefaultScramArch)

            # check for era or run dependent config parameters
            streamConfig.Express.GlobalTag = extractConfigParameter(streamConfig.Express.GlobalTag, runInfo['acq_era'], run)
            streamConfig.Express.ProcessingVersion = extractConfigParameter(streamConfig.Express.ProcessingVersion, runInfo['acq_era'], run)

            write_tiers = ','.join(streamConfig.Express.DataTiers)
            if not write_tiers:
                write_tiers = None

            bindsExpressConfig = { 'RUN' : run,
                                   'STREAM' : stream,
                                   'PROC_VER' : streamConfig.Express.ProcessingVersion,
                                   'WRITE_TIERS' : write_tiers,
                                   'WRITE_DQM' : streamConfig.Express.WriteDQM,
                                   'GLOBAL_TAG' : streamConfig.Express.GlobalTag,
                                   'MAX_RATE' : streamConfig.Express.MaxInputRate,
                                   'MAX_EVENTS' : streamConfig.Express.MaxInputEvents,
                                   'MAX_SIZE' : streamConfig.Express.MaxInputSize,
                                   'MAX_FILES' : streamConfig.Express.MaxInputFiles,
                                   'MAX_LATENCY' : streamConfig.Express.MaxLatency,
                                   'DQM_INTERVAL' : streamConfig.Express.PeriodicHarvestInterval,
                                   'CMSSW' : streamConfig.Express.CMSSWVersion,
                                   'SCRAM_ARCH' : streamConfig.Express.ScramArch,
                                   'RECO_CMSSW' : streamConfig.Express.RecoCMSSWVersion,
                                   'RECO_SCRAM_ARCH' : streamConfig.Express.RecoScramArch,
                                   'DATA_TYPE' : streamConfig.Express.DataType,
                                   'MULTICORE' : streamConfig.Express.Multicore,
                                   'ALCA_SKIM' : alcaSkim,
                                   'DQM_SEQ' : dqmSeq }

        #
        # then configure datasets
        #
        getStreamDatasetTriggersDAO = daoFactory(classname = "RunConfig.GetStreamDatasetTriggers")
        datasetTriggers = getStreamDatasetTriggersDAO.execute(run, stream, transaction = False)

        for dataset, paths in list(datasetTriggers.items()):

            datasetConfig = retrieveDatasetConfig(tier0Config, dataset)

            selectEvents = []
            for path in sorted(paths):
                selectEvents.append("%s:%s" % (path, runInfo['process']))

            if streamConfig.ProcessingStyle == "Bulk":

                outputModuleDetails.append( { 'dataTier' : "RAW",
                                              'eventContent' : "ALL",
                                              'selectEvents' : selectEvents,
                                              'primaryDataset' : dataset } )

                if datasetConfig.ArchivalNode or datasetConfig.TapeNode or datasetConfig.DiskNode or datasetConfig.DiskNodeReco:

                    bindsPhEDExConfig.append( { 'RUN' : run,
                                                'PRIMDS' : dataset,
                                                'ARCHIVAL_NODE' : datasetConfig.ArchivalNode,
                                                'TAPE_NODE' : datasetConfig.TapeNode,
                                                'DISK_NODE' : datasetConfig.DiskNode,
                                                'DISK_NODE_RECO' : datasetConfig.DiskNodeReco } )

                custodialSites = []
                nonCustodialSites = []
                autoApproveSites = []
                if datasetConfig.ArchivalNode:
                    bindsStorageNode.append( { 'NODE' : datasetConfig.ArchivalNode } )
                    custodialSites.append(datasetConfig.ArchivalNode)
                    autoApproveSites.append(datasetConfig.ArchivalNode)
                if datasetConfig.TapeNode:
                    bindsStorageNode.append( { 'NODE' : datasetConfig.TapeNode } )
                    if datasetConfig.RAWTapeNode:
                        bindsStorageNode.append( { 'NODE' : datasetConfig.RAWTapeNode } )
                        custodialSites.append(datasetConfig.RAWTapeNode)
                    else:
                        custodialSites.append(datasetConfig.TapeNode)
                if datasetConfig.DiskNode:
                    bindsStorageNode.append( { 'NODE' : datasetConfig.DiskNode } )
                    if datasetConfig.RAWtoDisk:
                        nonCustodialSites.append(datasetConfig.DiskNode)
                        autoApproveSites.append(datasetConfig.DiskNode)
                if datasetConfig.DiskNodeReco:
                    bindsStorageNode.append( { 'NODE' : datasetConfig.DiskNodeReco } )

                if len(custodialSites) > 0 or len(nonCustodialSites) > 0:
                    subscriptions.append( { 'custodialSites' : custodialSites,
                                            'custodialSubType' : "Replica",
                                            'custodialGroup' : "DataOps",
                                            'nonCustodialSites' : nonCustodialSites,
                                            'nonCustodialSubType' : "Replica",
                                            'nonCustodialGroup' : "AnalysisOps",
                                            'autoApproveSites' : autoApproveSites,
                                            'priority' : "high",
                                            'primaryDataset' : dataset,
                                            'deleteFromSource' : True,
                                            'dataTier' : "RAW" } )

                #
                # set subscriptions for error dataset
                #
                if datasetConfig.ArchivalNode != None:
                    subscriptions.append( { 'custodialSites' : [ datasetConfig.ArchivalNode ],
                                            'custodialSubType' : "Replica",
                                            'custodialGroup' : "DataOps",
                                            'autoApproveSites' : [ datasetConfig.ArchivalNode ],
                                            'priority' : "high",
                                            'primaryDataset' : "%s-Error" % dataset,
                                            'deleteFromSource' : True,
                                            'dataTier' : "RAW" } )


            elif streamConfig.ProcessingStyle == "Express":

                for dataTier in streamConfig.Express.DataTiers:

                    outputModuleDetails.append( { 'dataTier' : dataTier,
                                                  'eventContent' : dataTier,
                                                  'selectEvents' : selectEvents,
                                                  'primaryDataset' : dataset } )

                if streamConfig.Express.ArchivalNode or streamConfig.Express.TapeNode or streamConfig.Express.DiskNode:

                    bindsPhEDExConfig.append( { 'RUN' : run,
                                                'PRIMDS' : dataset,
                                                'ARCHIVAL_NODE' : streamConfig.Express.ArchivalNode,
                                                'TAPE_NODE' : streamConfig.Express.TapeNode,
                                                'DISK_NODE' : streamConfig.Express.DiskNode,
                                                'DISK_NODE_RECO' : None } )

                    custodialSites = []
                    nonCustodialSites = []
                    autoApproveSites = []
                    if streamConfig.Express.ArchivalNode:
                        custodialSites.append(streamConfig.Express.ArchivalNode)
                        autoApproveSites.append(streamConfig.Express.ArchivalNode)
                    if streamConfig.Express.TapeNode:
                        custodialSites.append(streamConfig.Express.TapeNode)
                    if streamConfig.Express.DiskNode:
                        nonCustodialSites.append(streamConfig.Express.DiskNode)
                        autoApproveSites.append(streamConfig.Express.DiskNode)

                    if len(custodialSites) > 0 or len(nonCustodialSites) > 0:
                        subscriptions.append( { 'custodialSites' : custodialSites,
                                                'custodialSubType' : "Replica",
                                                'custodialGroup' : streamConfig.Express.PhEDExGroup,
                                                'nonCustodialSites' : nonCustodialSites,
                                                'nonCustodialSubType' : "Replica",
                                                'nonCustodialGroup' : streamConfig.Express.PhEDExGroup,
                                                'autoApproveSites' : autoApproveSites,
                                                'priority' : "high",
                                                'primaryDataset' : dataset,
                                                'deleteFromSource' : True } )

        #
        # finally create WMSpec
        #
        outputs = {}
        if streamConfig.ProcessingStyle == "Bulk":

            taskName = "Repack"

            if tier0Config.Global.EnableUniqueWorkflowName:
                workflowName = "Repack_Run%d_Stream%s_%s_v%s_%s" % (run, stream, 
                    tier0Config.Global.AcquisitionEra, streamConfig.Repack.ProcessingVersion, 
                    time.strftime('%y%m%d_%H%M', time.localtime(time.time())))
            else:
                workflowName = "Repack_Run%d_Stream%s" % (run, stream)

            specArguments = {}

            specArguments['Memory'] = streamConfig.Repack.MaxMemory
            specArguments['Requestor'] = "Tier0"
            specArguments['RequestName'] = workflowName
            specArguments['RequestString'] = workflowName
            specArguments['RequestorDN'] = "Tier0"
            specArguments['RequestDate'] = []
            specArguments['RequestTransition'] = []
            specArguments['RequestStatus'] = REQUEST_START_STATE
            specArguments['RequestPriority'] = tier0Config.Global.BaseRequestPriority + 5000
            specArguments['PriorityTransition'] = []

            specArguments['CMSSWVersion'] = streamConfig.Repack.CMSSWVersion
            specArguments['ScramArch'] = streamConfig.Repack.ScramArch
            specArguments['ProcessingVersion'] = streamConfig.Repack.ProcessingVersion

            specArguments['MaxSizeSingleLumi'] = streamConfig.Repack.MaxSizeSingleLumi
            specArguments['MaxSizeMultiLumi'] = streamConfig.Repack.MaxSizeMultiLumi
            specArguments['MinInputSize'] = streamConfig.Repack.MinInputSize
            specArguments['MaxInputSize'] = streamConfig.Repack.MaxInputSize
            specArguments['MaxEdmSize'] = streamConfig.Repack.MaxEdmSize
            specArguments['MaxOverSize'] = streamConfig.Repack.MaxOverSize
            specArguments['MaxInputEvents'] = streamConfig.Repack.MaxInputEvents
            specArguments['MaxInputFiles'] = streamConfig.Repack.MaxInputFiles
            specArguments['MaxLatency'] = streamConfig.Repack.MaxLatency

            # parameters for repack direct to merge stageout
            specArguments['MinMergeSize'] = streamConfig.Repack.MinInputSize
            specArguments['MaxMergeEvents'] = streamConfig.Repack.MaxInputEvents

            specArguments['UnmergedLFNBase'] = "/store/unmerged/%s" % runInfo['bulk_data_type']
            if runInfo['backfill']:
                specArguments['MergedLFNBase'] = "/store/backfill/%s/%s" % (runInfo['backfill'],
                                                                            runInfo['bulk_data_type'])
            else:
                specArguments['MergedLFNBase'] = "/store/%s" % runInfo['bulk_data_type']

            blockCloseDelay = streamConfig.Repack.BlockCloseDelay

        elif streamConfig.ProcessingStyle == "Express":

            taskName = "Express"

            if tier0Config.Global.EnableUniqueWorkflowName:
                workflowName = "Express_Run%d_Stream%s_%s_v%s_%s" % (run, stream, 
                    tier0Config.Global.AcquisitionEra, streamConfig.Express.ProcessingVersion,
                    time.strftime('%y%m%d_%H%M', time.localtime(time.time())))
            else:
                workflowName = "Express_Run%d_Stream%s" % (run, stream)

            specArguments = {}

            specArguments['TimePerEvent'] = streamConfig.Express.TimePerEvent
            specArguments['SizePerEvent'] = streamConfig.Express.SizePerEvent

            if streamConfig.Express.Scenario == "HeavyIonsRun2":
                baseMemory = 3000
                perCoreMemory = 1300
            else:
                baseMemory = 3500
                perCoreMemory = 1050

            specArguments['Memory'] = baseMemory + perCoreMemory

            if streamConfig.Express.Multicore:
                specArguments['Multicore'] = streamConfig.Express.Multicore
                specArguments['Memory'] += (streamConfig.Express.Multicore - 1) * perCoreMemory

            specArguments['Requestor'] = "Tier0"
            specArguments['RequestName'] = workflowName
            specArguments['RequestString'] = workflowName
            specArguments['RequestorDN'] = "Tier0"
            specArguments['RequestDate'] = []
            specArguments['RequestTransition'] = []
            specArguments['RequestStatus'] = REQUEST_START_STATE
            specArguments['RequestPriority'] = tier0Config.Global.BaseRequestPriority + 10000
            specArguments['PriorityTransition'] = []

            specArguments['ProcessingString'] = "Express"
            specArguments['ProcessingVersion'] = streamConfig.Express.ProcessingVersion
            specArguments['Scenario'] = streamConfig.Express.Scenario

            specArguments['CMSSWVersion'] = streamConfig.Express.CMSSWVersion
            specArguments['ScramArch'] = streamConfig.Express.ScramArch
            specArguments['RecoCMSSWVersion'] = streamConfig.Express.RecoCMSSWVersion
            specArguments['RecoScramArch'] = streamConfig.Express.RecoScramArch

            specArguments['GlobalTag'] = streamConfig.Express.GlobalTag
            specArguments['GlobalTagTransaction'] = "Express_%d" % run
            specArguments['GlobalTagConnect'] = streamConfig.Express.GlobalTagConnect

            specArguments['MaxInputRate'] = streamConfig.Express.MaxInputRate
            specArguments['MaxInputEvents'] = streamConfig.Express.MaxInputEvents
            specArguments['MaxInputSize'] = streamConfig.Express.MaxInputSize
            specArguments['MaxInputFiles'] = streamConfig.Express.MaxInputFiles
            specArguments['MaxLatency'] = streamConfig.Express.MaxLatency
            specArguments['AlcaSkims'] = streamConfig.Express.AlcaSkims
            specArguments['DQMSequences'] = streamConfig.Express.DqmSequences
            specArguments['AlcaHarvestTimeout'] = runInfo['ah_timeout']
            specArguments['AlcaHarvestCondLFNBase'] = runInfo['ah_cond_lfnbase']
            specArguments['AlcaHarvestLumiURL'] = runInfo['ah_lumi_url']
            specArguments['DQMUploadProxy'] = dqmUploadProxy
            specArguments['DQMUploadUrl'] = runInfo['dqmuploadurl']
            specArguments['StreamName'] = stream
            specArguments['SpecialDataset'] = specialDataset

            specArguments['UnmergedLFNBase'] = "/store/unmerged/%s" % streamConfig.Express.DataType
            if runInfo['backfill']:
                specArguments['MergedLFNBase'] = "/store/backfill/%s/%s" % (runInfo['backfill'],
                                                                            streamConfig.Express.DataType)
            else:
                specArguments['MergedLFNBase'] = "/store/%s" % streamConfig.Express.DataType

            specArguments['PeriodicHarvestInterval'] = streamConfig.Express.PeriodicHarvestInterval

            blockCloseDelay = streamConfig.Express.BlockCloseDelay

        if streamConfig.ProcessingStyle in [ 'Bulk', 'Express' ]:

            specArguments['RunNumber'] = run
            specArguments['AcquisitionEra'] = runInfo['acq_era']
            specArguments['Outputs'] = outputModuleDetails
            specArguments['ValidStatus'] = "VALID"

        if streamConfig.ProcessingStyle == "Bulk":
            factory = RepackWorkloadFactory()
            wmSpec = factory.factoryWorkloadConstruction(workflowName, specArguments)
            for subscription in subscriptions:
                wmSpec.setSubscriptionInformation(**subscription)
        elif streamConfig.ProcessingStyle == "Express":
            factory = ExpressWorkloadFactory()
            wmSpec = factory.factoryWorkloadConstruction(workflowName, specArguments)
            for subscription in subscriptions:
                wmSpec.setSubscriptionInformation(**subscription)

        if streamConfig.ProcessingStyle in [ 'Bulk', 'Express' ]:
            wmSpec.setOwnerDetails("Dirk.Hufnagel@cern.ch", "T0",
                                   { 'vogroup': 'DEFAULT', 'vorole': 'DEFAULT',
                                     'dn' : "Dirk.Hufnagel@cern.ch" } )

            wmSpec.updateArguments( { 'SiteWhitelist': [ tier0Config.Global.ProcessingSite ],
                                      'SiteBlacklist': [],
                                      'BlockCloseMaxWaitTime': blockCloseDelay,
                                      'SoftTimeout': 604800, #7 days, effectively disabled
                                      'GracePeriod': 3600,
                                      'Dashboard': "t0" } )

            if tier0Config.Global.ProcessingSite!=tier0Config.Global.StorageSite:
                site = retrieveSiteConfig(tier0Config,tier0Config.Global.StorageSite)
                wmSpec.setTaskEnvironmentVariables({'WMAGENT_SITE_CONFIG_OVERRIDE':site.SiteLocalConfig})
                wmSpec.setOverrideCatalog(site.OverrideCatalog)

            wmbsHelper = WMBSHelper(wmSpec, taskName, cachepath = specDirectory)

        filesetName = "Run%d_Stream%s" % (run, stream)
        fileset = Fileset(filesetName)

        #
        # create workflow (currently either repack or express)
        #
        try:
            myThread.transaction.begin()
            if len(bindsCMSSWVersion) > 0:
                insertCMSSWVersionDAO.execute(bindsCMSSWVersion, conn = myThread.transaction.conn, transaction = True)
            if len(bindsDataset) > 0:
                insertDatasetDAO.execute(bindsDataset, conn = myThread.transaction.conn, transaction = True)
            if len(bindsStreamDataset) > 0:
                insertStreamDatasetDAO.execute(bindsStreamDataset, conn = myThread.transaction.conn, transaction = True)
            if len(bindsRepackConfig) > 0:
                insertRepackConfigDAO.execute(bindsRepackConfig, conn = myThread.transaction.conn, transaction = True)
            if len(bindsPromptCalibration) > 0:
                insertPromptCalibrationDAO.execute(bindsPromptCalibration, conn = myThread.transaction.conn, transaction = True)
            if len(bindsExpressConfig) > 0:
                insertExpressConfigDAO.execute(bindsExpressConfig, conn = myThread.transaction.conn, transaction = True)
            if len(bindsSpecialDataset) > 0:
                insertSpecialDatasetDAO.execute(bindsSpecialDataset, conn = myThread.transaction.conn, transaction = True)
            if len(bindsDatasetScenario) > 0:
                insertDatasetScenarioDAO.execute(bindsDatasetScenario, conn = myThread.transaction.conn, transaction = True)
            if len(bindsStorageNode) > 0:
                insertStorageNodeDAO.execute(bindsStorageNode, conn = myThread.transaction.conn, transaction = True)
            if len(bindsPhEDExConfig) > 0:
                insertPhEDExConfigDAO.execute(bindsPhEDExConfig, conn = myThread.transaction.conn, transaction = True)
            insertRunStreamDoneDAO.execute(bindsRunStreamDone, conn = myThread.transaction.conn, transaction = True)
            insertStreamStyleDAO.execute(bindsStreamStyle, conn = myThread.transaction.conn, transaction = True)
            if streamConfig.ProcessingStyle in [ 'Bulk', 'Express' ]:
                insertStreamFilesetDAO.execute(run, stream, filesetName, conn = myThread.transaction.conn, transaction = True)
                fileset.load()
                wmbsHelper.createSubscription(wmSpec.getTask(taskName), fileset, alternativeFilesetClose = True)
                insertWorkflowMonitoringDAO.execute([fileset.id],  conn = myThread.transaction.conn, transaction = True)
            if streamConfig.ProcessingStyle == "Bulk":
                bindsRecoReleaseConfig = []
                for fileset, primds in list(wmbsHelper.getMergeOutputMapping().items()):
                    bindsRecoReleaseConfig.append( { 'RUN' : run,
                                                     'PRIMDS' : primds,
                                                     'FILESET' : fileset } )
                insertRecoReleaseConfigDAO.execute(bindsRecoReleaseConfig, conn = myThread.transaction.conn, transaction = True)
        except Exception as ex:
            logging.exception(ex)
            myThread.transaction.rollback()
            raise RuntimeError("Problem in configureRunStream() database transaction !")
        else:
            myThread.transaction.commit()

    else:

        # should we do anything for local runs ?
        pass
    return

def releasePromptReco(tier0Config, specDirectory, dqmUploadProxy):
    """
    _releasePromptReco_

    Called by Tier0Feeder

    Finds all run/primds that need to be released for PromptReco
    ( run.stop_time + reco_release_config.delay > now AND run.stop_time > 0 )

    Create workflows and subscriptions for the processing
    of runs/datasets.

    """
    logging.debug("releasePromptReco()")
    myThread = threading.currentThread()

    daoFactory = DAOFactory(package = "T0.WMBS",
                            logger = logging,
                            dbinterface = myThread.dbi)

    findRecoReleaseDatasetsDAO = daoFactory(classname = "RunConfig.FindRecoReleaseDatasets")
    findRecoReleaseDAO = daoFactory(classname = "RunConfig.FindRecoRelease")
    insertDatasetScenarioDAO = daoFactory(classname = "RunConfig.InsertDatasetScenario")
    insertCMSSWVersionDAO = daoFactory(classname = "RunConfig.InsertCMSSWVersion")
    insertRecoConfigDAO = daoFactory(classname = "RunConfig.InsertRecoConfig")
    insertStorageNodeDAO = daoFactory(classname = "RunConfig.InsertStorageNode")
    insertPhEDExConfigDAO = daoFactory(classname = "RunConfig.InsertPhEDExConfig")
    releasePromptRecoDAO = daoFactory(classname = "RunConfig.ReleasePromptReco")
    insertWorkflowMonitoringDAO = daoFactory(classname = "RunConfig.InsertWorkflowMonitoring")

    # mark workflows as injected
    wmbsDaoFactory = DAOFactory(package = "WMCore.WMBS",
                                logger = logging,
                                dbinterface = myThread.dbi)
    markWorkflowsInjectedDAO   = wmbsDaoFactory(classname = "Workflow.MarkInjectedWorkflows")

    #
    # handle PromptReco release for datasets
    #
    recoReleaseDatasets = findRecoReleaseDatasetsDAO.execute(transaction = False)

    datasetDelays = {}
    for dataset in recoReleaseDatasets:
        datasetConfig = retrieveDatasetConfig(tier0Config, dataset)
        datasetDelays[dataset] = (datasetConfig.RecoDelay, datasetConfig.RecoDelayOffset)

    recoRelease = findRecoReleaseDAO.execute(datasetDelays, transaction = False)
    for run in sorted(recoRelease.keys()):

        # for creating PromptReco specs
        recoSpecs = {}

        # for PhEDEx subscription settings
        subscriptions = []

        bindsDatasetScenario = []
        bindsCMSSWVersion = []
        bindsRecoConfig = []
        bindsStorageNode = []
        bindsReleasePromptReco = []

        # retrieve some basic run information
        getRunInfoDAO = daoFactory(classname = "RunConfig.GetRunInfo")
        runInfo = getRunInfoDAO.execute(run, transaction = False)[0]

        # retrieve phedex configs for run
        getPhEDExConfigDAO = daoFactory(classname = "RunConfig.GetPhEDExConfig")
        phedexConfigs = getPhEDExConfigDAO.execute(run, transaction = False)

        for (dataset, fileset, repackProcVer) in recoRelease[run]:

            bindsReleasePromptReco.append( { 'RUN' : run,
                                             'PRIMDS' : dataset,
                                             'NOW' : int(time.time()) } )

            datasetConfig = retrieveDatasetConfig(tier0Config, dataset)

            # check for era or run dependent config parameters
            datasetConfig.Scenario = extractConfigParameter(datasetConfig.Scenario, runInfo['acq_era'], run)
            datasetConfig.CMSSWVersion = extractConfigParameter(datasetConfig.CMSSWVersion, runInfo['acq_era'], run)
            datasetConfig.GlobalTag = extractConfigParameter(datasetConfig.GlobalTag, runInfo['acq_era'], run)
            datasetConfig.ProcessingVersion = extractConfigParameter(datasetConfig.ProcessingVersion, runInfo['acq_era'], run)

            bindsDatasetScenario.append( { 'RUN' : run,
                                           'PRIMDS' : dataset,
                                           'SCENARIO' : datasetConfig.Scenario } )

            bindsCMSSWVersion.append( { 'VERSION' : datasetConfig.CMSSWVersion } )

            alcaSkim = None
            if len(datasetConfig.AlcaSkims) > 0:
                alcaSkim = ",".join(datasetConfig.AlcaSkims)

            physicsSkim = None
            if len(datasetConfig.PhysicsSkims) > 0:
                physicsSkim = ",".join(datasetConfig.PhysicsSkims)

            dqmSeq = None
            if len(datasetConfig.DqmSequences) > 0:
                dqmSeq = ",".join(datasetConfig.DqmSequences)

            datasetConfig.ScramArch = tier0Config.Global.ScramArches.get(datasetConfig.CMSSWVersion,
                                                                         tier0Config.Global.DefaultScramArch)

            bindsRecoConfig.append( { 'RUN' : run,
                                      'PRIMDS' : dataset,
                                      'DO_RECO' : int(datasetConfig.DoReco),
                                      'RECO_SPLIT' : datasetConfig.RecoSplit,
                                      'WRITE_RECO' : int(datasetConfig.WriteRECO),
                                      'WRITE_DQM' : int(datasetConfig.WriteDQM),
                                      'WRITE_AOD' : int(datasetConfig.WriteAOD),
                                      'WRITE_MINIAOD' : int(datasetConfig.WriteMINIAOD),
                                      'PROC_VER' : datasetConfig.ProcessingVersion,
                                      'ALCA_SKIM' : alcaSkim,
                                      'PHYSICS_SKIM' : physicsSkim,
                                      'DQM_SEQ' : dqmSeq,
                                      'CMSSW' : datasetConfig.CMSSWVersion,
                                      'SCRAM_ARCH' : datasetConfig.ScramArch,
                                      'MULTICORE' : datasetConfig.Multicore,
                                      'GLOBAL_TAG' : datasetConfig.GlobalTag } )

            # check if the dataset has any phedex config
            if dataset in phedexConfigs:

                phedexConfig = phedexConfigs[dataset]

                tapeDataTiers = set()
                diskDataTiers = set()
                skimDataTiers = set()
                alcaDataTiers = set()

                if datasetConfig.WriteRECO:
                    diskDataTiers.add("RECO")
                if datasetConfig.WriteAOD:
                    tapeDataTiers.add("AOD")
                    diskDataTiers.add("AOD")
                if datasetConfig.WriteMINIAOD:
                    tapeDataTiers.add("MINIAOD")
                    diskDataTiers.add("MINIAOD")
                if datasetConfig.WriteDQM:
                    tapeDataTiers.add(tier0Config.Global.DQMDataTier)
                if len(datasetConfig.PhysicsSkims) > 0:
                    skimDataTiers.add("RAW-RECO")
                    skimDataTiers.add("USER")
                    skimDataTiers.add("RECO")
                    skimDataTiers.add("AOD")
                if len(datasetConfig.AlcaSkims) > 0:
                    alcaDataTiers.add("ALCARECO")

                # do things different based on whether we have TapeNode/DiskNode, only TapeNode or ArchivalNode
                if phedexConfig['tape_node'] != None:

                    if phedexConfig['disk_node'] == None:
                        diskDataTiers = set()

                    for dataTier in tapeDataTiers & diskDataTiers:

                        if dataTier == "RECO" and phedexConfig['disk_node_reco']:
                            diskNode = phedexConfig['disk_node_reco']
                        else:
                            diskNode = phedexConfig['disk_node']

                        subscriptions.append( { 'custodialSites' : [phedexConfig['tape_node']],
                                                'custodialSubType' : "Replica",
                                                'custodialGroup' : "DataOps",
                                                'nonCustodialSites' : [diskNode],
                                                'nonCustodialSubType' : "Replica",
                                                'nonCustodialGroup' : "AnalysisOps",
                                                'autoApproveSites' : [diskNode],
                                                'priority' : "high",
                                                'primaryDataset' : dataset,
                                                'useSkim' : True,
                                                'isSkim' : False,
                                                'deleteFromSource' : True,
                                                'dataTier' : dataTier } )

                    for dataTier in skimDataTiers:
                        subscriptions.append( { 'custodialSites' : [phedexConfig['tape_node']],
                                                'custodialSubType' : "Replica",
                                                'custodialGroup' : "DataOps",
                                                'nonCustodialSites' : [phedexConfig['disk_node']] if phedexConfig['disk_node'] else [],
                                                'nonCustodialSubType' : "Replica",
                                                'nonCustodialGroup' : "AnalysisOps",
                                                'autoApproveSites' : [phedexConfig['disk_node']] if phedexConfig['disk_node'] else [],
                                                'priority' : "high",
                                                'primaryDataset' : dataset,
                                                'useSkim' : True,
                                                'isSkim' : True,
                                                'deleteFromSource' : True,
                                                'dataTier' : dataTier } )

                    for dataTier in tapeDataTiers - diskDataTiers:
                        subscriptions.append( { 'custodialSites' : [phedexConfig['tape_node']],
                                                'custodialSubType' : "Replica",
                                                'custodialGroup' : "DataOps",
                                                'autoApproveSites' : [],
                                                'priority' : "high",
                                                'primaryDataset' : dataset,
                                                'useSkim' : True,
                                                'isSkim' : False,
                                                'deleteFromSource' : True,
                                                'dataTier' : dataTier } )

                    for dataTier in alcaDataTiers:
                        subscriptions.append( { 'custodialSites' : [phedexConfig['tape_node']],
                                                'custodialSubType' : "Replica",
                                                'custodialGroup' : "DataOps",
                                                'autoApproveSites' : [],
                                                'priority' : "high",
                                                'primaryDataset' : dataset,
                                                'useSkim' : True,
                                                'isSkim' : True,
                                                'deleteFromSource' : True,
                                                'dataTier' : dataTier } )

                    for dataTier in diskDataTiers - tapeDataTiers:

                        if dataTier == "RECO" and phedexConfig['disk_node_reco']:
                            diskNode = phedexConfig['disk_node_reco']
                        else:
                            diskNode = phedexConfig['disk_node']

                        subscriptions.append( { 'nonCustodialSites' : [diskNode],
                                                'nonCustodialSubType' : "Replica",
                                                'nonCustodialGroup' : "AnalysisOps",
                                                'autoApproveSites' : [diskNode],
                                                'priority' : "high",
                                                'primaryDataset' : dataset,
                                                'useSkim' : True,
                                                'isSkim' : False,
                                                'deleteFromSource' : True,
                                                'dataTier' : dataTier } )

                elif phedexConfig['archival_node'] != None:

                    for dataTier in tapeDataTiers | diskDataTiers | skimDataTiers | alcaDataTiers:

                        subscriptions.append( { 'custodialSites' : [phedexConfig['archival_node']],
                                                'custodialSubType' : "Replica",
                                                'custodialGroup' : "DataOps",
                                                'autoApproveSites' : [phedexConfig['archival_node']],
                                                'priority' : "high",
                                                'primaryDataset' : dataset,
                                                'deleteFromSource' : True,
                                                'dataTier' : dataTier } )

            writeTiers = []
            if datasetConfig.WriteRECO:
                writeTiers.append("RECO")
            if datasetConfig.WriteAOD:
                writeTiers.append("AOD")
            if datasetConfig.WriteMINIAOD:
                writeTiers.append("MINIAOD")
            if datasetConfig.WriteDQM:
                writeTiers.append(tier0Config.Global.DQMDataTier)
            if len(datasetConfig.AlcaSkims) > 0:
                writeTiers.append("ALCARECO")

            if datasetConfig.DoReco and len(writeTiers) > 0:

                #
                # create WMSpec
                #
                taskName = "Reco"

                if tier0Config.Global.EnableUniqueWorkflowName:
                    workflowName = "PromptReco_Run%d_%s_%s_v%s_%s" % (run, dataset,
                        tier0Config.Global.AcquisitionEra, datasetConfig.ProcessingVersion,
                        time.strftime('%y%m%d_%H%M', time.localtime(time.time())))
                else:
                    workflowName = "PromptReco_Run%d_%s" % (run, dataset)

                specArguments = {}

                specArguments['TimePerEvent'] = datasetConfig.TimePerEvent
                specArguments['SizePerEvent'] = datasetConfig.SizePerEvent

                if datasetConfig.Scenario == "HeavyIonsRun2":
                    baseMemory = 3000
                    perCoreMemory = 1300
                else:
                    baseMemory = 3500
                    perCoreMemory = 1050

                specArguments['Memory'] = baseMemory + perCoreMemory

                if datasetConfig.Multicore:
                    specArguments['Multicore'] = datasetConfig.Multicore
                    specArguments['Memory'] += (datasetConfig.Multicore - 1) * perCoreMemory

                specArguments['Memory'] += len(datasetConfig.PhysicsSkims) * 100

                specArguments['Requestor'] = "Tier0"
                specArguments['RequestName'] = workflowName
                specArguments['RequestString'] = workflowName
                specArguments['RequestorDN'] = "Tier0"
                specArguments['RequestDate'] = []
                specArguments['RequestTransition'] = []
                specArguments['RequestStatus'] = REQUEST_START_STATE
                specArguments['RequestPriority'] = tier0Config.Global.BaseRequestPriority
                specArguments['PriorityTransition'] = []

                specArguments['AcquisitionEra'] = runInfo['acq_era']
                specArguments['CMSSWVersion'] = datasetConfig.CMSSWVersion
                specArguments['ScramArch'] = datasetConfig.ScramArch

                specArguments['RunNumber'] = run

                specArguments['SplittingAlgo'] = "EventAwareLumiBased"
                specArguments['EventsPerJob'] = datasetConfig.RecoSplit

                specArguments['RobustMerge'] = False

                specArguments['ProcessingString'] = "PromptReco"
                specArguments['ProcessingVersion'] = datasetConfig.ProcessingVersion
                specArguments['Scenario'] = datasetConfig.Scenario

                specArguments['GlobalTag'] = datasetConfig.GlobalTag
                specArguments['GlobalTagConnect'] = datasetConfig.GlobalTagConnect

                specArguments['InputDataset'] = "/%s/%s-%s/RAW" % (dataset, runInfo['acq_era'], repackProcVer)

                specArguments['WriteTiers'] = writeTiers
                specArguments['AlcaSkims'] = datasetConfig.AlcaSkims
                specArguments['PhysicsSkims'] = datasetConfig.PhysicsSkims
                specArguments['DQMSequences'] = datasetConfig.DqmSequences

                specArguments['UnmergedLFNBase'] = "/store/unmerged/%s" % runInfo['bulk_data_type']
                if runInfo['backfill']:
                    specArguments['MergedLFNBase'] = "/store/backfill/%s/%s" % (runInfo['backfill'],
                                                                                runInfo['bulk_data_type'])
                else:
                    specArguments['MergedLFNBase'] = "/store/%s" % runInfo['bulk_data_type']

                specArguments['ValidStatus'] = "VALID"

                specArguments['EnableHarvesting'] = "True"
                specArguments['DQMUploadProxy'] = dqmUploadProxy
                specArguments['DQMUploadUrl'] = runInfo['dqmuploadurl']

                factory = PromptRecoWorkloadFactory()
                wmSpec = factory.factoryWorkloadConstruction(workflowName, specArguments)
                for subscription in subscriptions:
                    wmSpec.setSubscriptionInformation(**subscription)

                wmSpec.setOwnerDetails("Dirk.Hufnagel@cern.ch", "T0",
                                       { 'vogroup': 'DEFAULT', 'vorole': 'DEFAULT',
                                         'dn' : "Dirk.Hufnagel@cern.ch" } )

                #Overriding site configuration
                if tier0Config.Global.ProcessingSite!=tier0Config.Global.StorageSite:
                    site = retrieveSiteConfig(tier0Config,tier0Config.Global.StorageSite)
                    wmSpec.setTaskEnvironmentVariables({'WMAGENT_SITE_CONFIG_OVERRIDE':site.SiteLocalConfig})
                    wmSpec.setOverrideCatalog(site.OverrideCatalog)

                #Overriding processing site in case we using T0 disk
                wmSpec.updateArguments( { 'SiteWhitelist': datasetConfig.SiteWhitelist,
                                          'SiteBlacklist': [],
                                          'TrustSitelists': "True",
                                          'BlockCloseMaxWaitTime': datasetConfig.BlockCloseDelay,
                                          'SoftTimeout': 604800, #7 days, effectively disabled
                                          'GracePeriod': 3600,
                                          'Dashboard': "t0" } )

                wmbsHelper = WMBSHelper(wmSpec, taskName, cachepath = specDirectory)

                recoSpecs[workflowName] = (wmbsHelper, wmSpec, fileset)

        try:
            myThread.transaction.begin()
            if len(bindsDatasetScenario) > 0:
                insertDatasetScenarioDAO.execute(bindsDatasetScenario, conn = myThread.transaction.conn, transaction = True)
            if len(bindsCMSSWVersion) > 0:
                insertCMSSWVersionDAO.execute(bindsCMSSWVersion, conn = myThread.transaction.conn, transaction = True)
            if len(bindsRecoConfig) > 0:
                insertRecoConfigDAO.execute(bindsRecoConfig, conn = myThread.transaction.conn, transaction = True)
            if len(bindsStorageNode) > 0:
                insertStorageNodeDAO.execute(bindsStorageNode, conn = myThread.transaction.conn, transaction = True)
            if len(bindsReleasePromptReco) > 0:
                releasePromptRecoDAO.execute(bindsReleasePromptReco, conn = myThread.transaction.conn, transaction = True)
            for (wmbsHelper, wmSpec, fileset) in list(recoSpecs.values()):
                wmbsHelper.createSubscription(wmSpec.getTask(taskName), Fileset(id = fileset), alternativeFilesetClose = True)
                insertWorkflowMonitoringDAO.execute([fileset],  conn = myThread.transaction.conn, transaction = True)
            if len(recoSpecs) > 0:
                markWorkflowsInjectedDAO.execute(list(recoSpecs.keys()), injected = True, conn = myThread.transaction.conn, transaction = True)
        except Exception as ex:
            logging.exception(ex)
            myThread.transaction.rollback()
            raise RuntimeError("Problem in releasePromptReco() database transaction !")
        else:
            myThread.transaction.commit()

    return



"""
_InsertRunStreamDone_

Oracle implementation of InsertRunStreamDone

Insert RunStreamDone record

"""

from WMCore.Database.DBFormatter import DBFormatter

class InsertRunStreamDone(DBFormatter):

    def execute(self, binds, conn = None, transaction = False):

        sql = """INSERT INTO run_stream_done
                 (run_id, stream_id)
                 VALUES (:RUN,
                         (SELECT id FROM stream WHERE name = :STREAM))
                 """

        self.dbi.processData(sql, binds, conn = conn,
                             transaction = transaction)

        return



"""
_MarkCloseoutWorkflowMonitoring_

Oracle implementation of MarkCloseoutWorkflowMonitoring
Sets a workflow as "closeout=1" in workflow_monitoring table, should be used after successful upload to couchDB
"""
from WMCore.Database.DBFormatter import DBFormatter

class MarkCloseoutWorkflowMonitoring(DBFormatter):

    def execute(self, workflowId, conn = None, transaction = False):

        sql = """UPDATE workflow_monitoring
                 SET closeout = 1
                 WHERE workflow = :WORKFLOW_ID"""

        binds = [{'WORKFLOW_ID': workflowId}]
        self.dbi.processData(sql, binds, conn = conn,
                             transaction = transaction)
        return 



"""
_GetFinishedStreamers_

Oracle implementation of GetFinishedStreamers

Returns all streamer files that are completely processed (file is complete or failed).

Query works as-is since streamer files are only in a single fileset/subscription.

"""

from WMCore.Database.DBFormatter import DBFormatter

class GetFinishedStreamers(DBFormatter):

    def execute(self, conn = None, transaction = False):

        sql = """SELECT streamer.id,
                        streamer.p5_id,
                        wmbs_file_details.lfn
                 FROM streamer
                 LEFT OUTER JOIN wmbs_sub_files_complete ON
                   wmbs_sub_files_complete.fileid = streamer.id
                 LEFT OUTER JOIN wmbs_sub_files_failed ON
                   wmbs_sub_files_failed.fileid = streamer.id
                 INNER JOIN wmbs_file_details ON
                   wmbs_file_details.id = streamer.id
                 WHERE checkForZeroState(streamer.deleted) = 0
                 AND ( wmbs_sub_files_complete.fileid IS NOT NULL
                       OR wmbs_sub_files_failed.fileid IS NOT NULL )
                 """

        results = self.dbi.processData(sql, {}, conn = conn,
                                       transaction = transaction)[0].fetchall()

        streamers = []
        for result in results:
            streamers.append( (result[0], result[1]) )

        return streamers



"""
_FindClosedLumis_

Oracle implementation of FindClosedLumis

Checking for closed run/stream/lumi with lumis above
a minimum threshold, extracting filecounts as well.

A lumi is considered closed if we have EoLS records
from every SM instance OR if we have EoLS records
from every SM instance except the ones that are not
supposed to have EoLS records anymore (because
the instance already has a EoR record with a max
lumi that is smaller than the current lumi).

Return a list of dictionaries with run/stream/lumi.

"""

from WMCore.Database.DBFormatter import DBFormatter

import logging
from sqlalchemy.exc import DatabaseError

class FindClosedLumis(DBFormatter):

    def execute(self, binds, conn = None, transaction = False):

        # query has problems because of the way the join work
        # can create two groups for the same run/stream/lumi
        # first counting comparison doesn't work then
##         sql = """SELECT b.runnumber, a.stream, a.lumisection, SUM(a.filecount)
##                  FROM CMS_STOMGR.runs b
##                  LEFT OUTER JOIN CMS_STOMGR.streams a ON
##                    a.runnumber = b.runnumber AND
##                    a.instance = b.instance AND
##                    a.stream = :STREAM AND
##                    a.lumisection > :LUMI
##                  WHERE b.runnumber = :RUN
##                  GROUP BY b.runnumber, a.stream, a.lumisection
##                  HAVING COUNT(b.runnumber) = MAX(b.n_instances)
##                  AND (
##                    COUNT(a.runnumber) = COUNT(b.runnumber) OR
##                    COUNT(a.runnumber) = SUM(CASE
##                                               WHEN b.status = 1 THEN 1000
##                                               WHEN b.n_lumisections < a.lumisection THEN 0
##                                               ELSE 1
##                                             END)
##                  )
##                  """

        sql = """SELECT a.runnumber
                 FROM CMS_STOMGR.runs a
                 WHERE a.runnumber = :RUN
                 GROUP BY a.runnumber
                 HAVING COUNT(*) = MAX(a.n_instances)
                 """

        bindVars = []
        for b in binds:
            bindVars.append( { 'RUN' : b['RUN'] } )

        results = self.dbi.processData(sql, bindVars, conn = conn,
                                       transaction = transaction)[0].fetchall()

        goodRuns = []
        for result in results:
            goodRuns.append(result[0])

        bindVars = []
        for b in binds:
            if b['RUN'] in goodRuns:
                bindVars.append(b)

        if len(bindVars) == 0:
            return []

        sql = """SELECT a.runnumber, a.stream, a.lumisection, SUM(a.filecount)
                 FROM CMS_STOMGR.streams a
                 INNER JOIN CMS_STOMGR.runs b ON
                   b.runnumber = a.runnumber AND
                   b.instance = a.instance
                 WHERE a.runnumber = :RUN
                 AND a.stream = :STREAM
                 AND a.lumisection > :LUMI
                 GROUP BY a.runnumber, a.stream, a.lumisection
                 HAVING ( COUNT(*) = ( SELECT COUNT(*)
                                       FROM CMS_STOMGR.runs c
                                       WHERE c.runnumber = a.runnumber ) AND
                          COUNT(*) = ( SELECT MAX(c.n_instances)
                                       FROM CMS_STOMGR.runs c
                                       WHERE c.runnumber = a.runnumber ) )
                 OR ( COUNT(*) = SUM(CASE
                                       WHEN b.status = 1 THEN 1000
                                       WHEN b.n_lumisections < a.lumisection THEN 0
                                       ELSE 1
                                     END) AND
                      COUNT(*) = ( SELECT SUM(CASE
                                                WHEN c.status = 1 THEN 1000
                                                WHEN c.n_lumisections < a.lumisection THEN 0
                                                ELSE 1
                                              END)
                                          FROM CMS_STOMGR.runs c
                                          WHERE c.runnumber = a.runnumber ) )
                 """

        try:
            results = self.dbi.processData(sql, bindVars, conn = conn,
                                           transaction = transaction)[0].fetchall()
        except DatabaseError as ex:
            logging.error("ERROR: DatabaseError exception when checking for closed run/stream/lumi")
            logging.error("   %s" % ex)
            results = []

        closedLumis = []
        for result in results:
            closedLumis.append( { 'RUN' : result[0],
                                  'STREAM' : result[1],
                                  'LUMI' : result[2],
                                  'FILECOUNT' : result[3] } )

        return closedLumis



"""
_OfflineConfiguration_

Processing configuration for the Tier0 - Replay version
"""
from __future__ import print_function

from T0.RunConfig.Tier0Config import addDataset
from T0.RunConfig.Tier0Config import createTier0Config
from T0.RunConfig.Tier0Config import setAcquisitionEra
from T0.RunConfig.Tier0Config import setScramArch
from T0.RunConfig.Tier0Config import setDefaultScramArch
from T0.RunConfig.Tier0Config import setBaseRequestPriority
from T0.RunConfig.Tier0Config import setBackfill
from T0.RunConfig.Tier0Config import setBulkDataType
from T0.RunConfig.Tier0Config import setProcessingSite
from T0.RunConfig.Tier0Config import setDQMDataTier
from T0.RunConfig.Tier0Config import setDQMUploadUrl
from T0.RunConfig.Tier0Config import setPromptCalibrationConfig
from T0.RunConfig.Tier0Config import setConfigVersion
from T0.RunConfig.Tier0Config import ignoreStream
from T0.RunConfig.Tier0Config import addRepackConfig
from T0.RunConfig.Tier0Config import addExpressConfig
from T0.RunConfig.Tier0Config import addRegistrationConfig
from T0.RunConfig.Tier0Config import addConversionConfig
from T0.RunConfig.Tier0Config import setInjectRuns
from T0.RunConfig.Tier0Config import setStreamerPNN
from T0.RunConfig.Tier0Config import setEnableUniqueWorkflowName

# Create the Tier0 configuration object
tier0Config = createTier0Config()

# Set the verstion configuration (not used at the moment)
setConfigVersion(tier0Config, "replace with real version")

# Set run number to replay
setInjectRuns(tier0Config, [ 326607 ])

# Settings up sites
processingSite = "T2_CH_CERN"
streamerPNN = "T2_CH_CERN"

# Set global parameters:
#  Acquisition era
#  BaseRequestPriority
#  Backfill mode
#  Data type
#  Processing site (where jobs run)
#  PhEDEx locations
setAcquisitionEra(tier0Config, "Tier0_REPLAY_vocms015")
setBaseRequestPriority(tier0Config, 251000)
setBackfill(tier0Config, 1)
setBulkDataType(tier0Config, "data")
setProcessingSite(tier0Config, processingSite)
setStreamerPNN(tier0Config, streamerPNN)

# Override for DQM data tier
setDQMDataTier(tier0Config, "DQMIO")

# Set unique replay workflow names
# Uses era name, Repack, Express, PromptReco processing versions, date/time, e.g.:
# PromptReco_Run322057_Charmonium_Tier0_REPLAY_vocms047_v274_190221_121
setEnableUniqueWorkflowName(tier0Config)

# Define the two default timeouts for reco release
# First timeout is used directly for reco release
# Second timeout is used for the data service PromptReco start check
# (to basically say we started PromptReco even though we haven't)
defaultRecoTimeout =  10 * 60
defaultRecoLockTimeout = 5 * 60

# DQM Server
setDQMUploadUrl(tier0Config, "https://cmsweb.cern.ch/dqm/dev;https://cmsweb-testbed.cern.ch/dqm/offline-test")

# PCL parameters
setPromptCalibrationConfig(tier0Config,
                           alcaHarvestTimeout = 12*3600,
                           alcaHarvestCondLFNBase = "/store/unmerged/tier0_harvest",
                           alcaHarvestLumiURL = "root://eoscms.cern.ch//eos/cms/store/unmerged/tier0_harvest",
                           conditionUploadTimeout = 18*3600,
                           dropboxHost = "webcondvm.cern.ch",
                           validationMode = True)

# Special syntax supported for cmssw version, processing version and global tag
#
# { 'acqEra': {'Era1': Value1, 'Era2': Value2},
#   'maxRun': {100000: Value3, 200000: Value4},
#   'default': Value5 }

# Defaults for CMSSW version
defaultCMSSWVersion = {
       'default': "CMSSW_10_3_1_patch2"
     }

# Configure ScramArch
setDefaultScramArch(tier0Config, "slc6_amd64_gcc700")

# Configure scenarios
ppScenario = "ppEra_Run2_2018"
ppScenarioB0T = "ppEra_Run2_2018"
cosmicsScenario = "cosmicsEra_Run2_2018"
hcalnzsScenario="hcalnzsEra_Run2_2018_pp_on_AA"
hiScenario = "ppEra_Run2_2016_pA"
alcaTrackingOnlyScenario = "trackingOnlyEra_Run2_2018"
HIalcaTrackingOnlyScenario = "trackingOnlyEra_Run2_2018_pp_on_AA"
alcaTestEnableScenario = "AlCaTestEnable"
alcaLumiPixelsScenario = "AlCaLumiPixels"
hiTestppScenario = "ppEra_Run2_2018_pp_on_AA"

# Defaults for processing version
defaultProcVersion = 221
expressProcVersion = 221
alcarawProcVersion = 221

# Defaults for GlobalTag
expressGlobalTag = "103X_dataRun2_Express_v2"
promptrecoGlobalTag = "103X_dataRun2_Prompt_v3"
alcap0GlobalTag = "103X_dataRun2_Prompt_v3"

# Mandatory for CondDBv2
globalTagConnect = "frontier://PromptProd/CMS_CONDITIONS"

# Multicore settings
numberOfCores = 8

# Splitting parameters for PromptReco
defaultRecoSplitting = 750 * numberOfCores
hiRecoSplitting = 200 * numberOfCores
alcarawSplitting = 20000 * numberOfCores

#
# Setup repack and express mappings
#
repackVersionOverride = {
    "CMSSW_10_0_0" : "CMSSW_10_3_1_patch2",
    "CMSSW_10_0_1" : "CMSSW_10_3_1_patch2",
    "CMSSW_10_0_2" : "CMSSW_10_3_1_patch2",
    "CMSSW_10_0_3" : "CMSSW_10_3_1_patch2",
    "CMSSW_10_0_4" : "CMSSW_10_3_1_patch2",
    "CMSSW_10_0_5" : "CMSSW_10_3_1_patch2",
    "CMSSW_10_1_0" : "CMSSW_10_3_1_patch2",
    "CMSSW_10_1_1" : "CMSSW_10_3_1_patch2",
    "CMSSW_10_1_2" : "CMSSW_10_3_1_patch2",
    "CMSSW_10_1_3" : "CMSSW_10_3_1_patch2",
    "CMSSW_10_1_4" : "CMSSW_10_3_1_patch2",
    "CMSSW_10_1_5" : "CMSSW_10_3_1_patch2",
    "CMSSW_10_1_6" : "CMSSW_10_3_1_patch2",
    "CMSSW_10_1_7" : "CMSSW_10_3_1_patch2",
    "CMSSW_10_1_8" : "CMSSW_10_3_1_patch2",
    "CMSSW_10_1_9" : "CMSSW_10_3_1_patch2",
    "CMSSW_10_1_10" : "CMSSW_10_3_1_patch2",
    "CMSSW_10_2_0" : "CMSSW_10_3_1_patch2",
    "CMSSW_10_2_1" : "CMSSW_10_3_1_patch2",
    "CMSSW_10_2_5" : "CMSSW_10_3_1_patch2",
    "CMSSW_10_3_0" : "CMSSW_10_3_1_patch2"
    }

expressVersionOverride = {
    "CMSSW_10_0_0" : "CMSSW_10_3_1_patch2",
    "CMSSW_10_0_1" : "CMSSW_10_3_1_patch2",
    "CMSSW_10_0_2" : "CMSSW_10_3_1_patch2",
    "CMSSW_10_0_3" : "CMSSW_10_3_1_patch2",
    "CMSSW_10_0_4" : "CMSSW_10_3_1_patch2",
    "CMSSW_10_0_5" : "CMSSW_10_3_1_patch2",
    "CMSSW_10_1_0" : "CMSSW_10_3_1_patch2",
    "CMSSW_10_1_1" : "CMSSW_10_3_1_patch2",
    "CMSSW_10_1_2" : "CMSSW_10_3_1_patch2",
    "CMSSW_10_1_3" : "CMSSW_10_3_1_patch2",
    "CMSSW_10_1_4" : "CMSSW_10_3_1_patch2",
    "CMSSW_10_1_5" : "CMSSW_10_3_1_patch2",
    "CMSSW_10_1_6" : "CMSSW_10_3_1_patch2",
    "CMSSW_10_1_7" : "CMSSW_10_3_1_patch2",
    "CMSSW_10_1_8" : "CMSSW_10_3_1_patch2",
    "CMSSW_10_1_9" : "CMSSW_10_3_1_patch2",
    "CMSSW_10_1_10" : "CMSSW_10_3_1_patch2",
    "CMSSW_10_2_0" : "CMSSW_10_3_1_patch2",
    "CMSSW_10_2_1" : "CMSSW_10_3_1_patch2",
    "CMSSW_10_2_5" : "CMSSW_10_3_1_patch2",
    "CMSSW_10_3_0" : "CMSSW_10_3_1_patch2"
    }

#set default repack settings for bulk streams
addRepackConfig(tier0Config, "Default",
                proc_ver = defaultProcVersion,
                maxSizeSingleLumi = 24 * 1024 * 1024 * 1024,
                maxSizeMultiLumi = 8 * 1024 * 1024 * 1024,
                minInputSize =  2.1 * 1024 * 1024 * 1024,
                maxInputSize = 4 * 1024 * 1024 * 1024,
                maxEdmSize = 24 * 1024 * 1024 * 1024,
                maxOverSize = 8 * 1024 * 1024 * 1024,
                maxInputEvents = 3 * 1000 * 1000,
                maxInputFiles = 1000,
                maxLatency = 24 * 3600,
                blockCloseDelay = 1200,
                versionOverride = repackVersionOverride)

addDataset(tier0Config, "Default",
           do_reco = False,
           write_reco = False, write_aod = True, write_miniaod = True, write_dqm = False,
           reco_delay = defaultRecoTimeout,
           reco_delay_offset = defaultRecoLockTimeout,
           reco_split = defaultRecoSplitting,
           proc_version = defaultProcVersion,
           cmssw_version = defaultCMSSWVersion,
           multicore = numberOfCores,
           global_tag = promptrecoGlobalTag,
           global_tag_connect = globalTagConnect,
#           archival_node = "T0_CH_CERN_MSS",
#           tape_node = "T1_US_FNAL_MSS",
#           disk_node = "T1_US_FNAL_Disk",
#           raw_to_disk = False,
           blockCloseDelay = 1200,
           timePerEvent = 5,
           sizePerEvent = 1500,
           scenario = ppScenario)

#############################
### Express configuration ###
#############################

addExpressConfig(tier0Config, "Express",
                 scenario = ppScenario,
                 data_tiers = [ "FEVT" ],
                 write_dqm = True,
                 alca_producers = [ "SiStripPCLHistos", "SiStripCalZeroBias", "SiStripCalMinBias", "SiStripCalMinBiasAAG",
                                    "TkAlMinBias", "DtCalib", "LumiPixelsMinBias", "SiPixelCalZeroBias",
                                    "PromptCalibProd", "PromptCalibProdSiStrip", "PromptCalibProdSiPixelAli",
                                    "PromptCalibProdSiStripGains", "PromptCalibProdSiStripGainsAAG", "PromptCalibProdSiPixel"
                                    ],
                 reco_version = defaultCMSSWVersion,
                 multicore = numberOfCores,
                 global_tag_connect = globalTagConnect,
                 global_tag = expressGlobalTag,
                 proc_ver = expressProcVersion,
                 maxInputRate = 23 * 1000,
                 maxInputEvents = 400,
                 maxInputSize = 2 * 1024 * 1024 * 1024,
                 maxInputFiles = 15,
                 maxLatency = 15 * 23,
                 periodicHarvestInterval = 20 * 60,
                 blockCloseDelay = 1200,
                 timePerEvent = 4,
                 sizePerEvent = 1700,
                 versionOverride = expressVersionOverride)

addExpressConfig(tier0Config, "ExpressCosmics",
                 scenario = cosmicsScenario,
                 data_tiers = [ "FEVT" ],
                 write_dqm = True,
                 alca_producers = [ "SiStripPCLHistos", "SiStripCalZeroBias", "TkAlCosmics0T",
                                    "DtCalibCosmics", "SiPixelCalZeroBias",
                                    "PromptCalibProdSiStrip", "PromptCalibProdSiPixel"
                                    ],
                 reco_version = defaultCMSSWVersion,
                 multicore = numberOfCores,
                 global_tag_connect = globalTagConnect,
                 global_tag = expressGlobalTag,
                 proc_ver = expressProcVersion,
                 maxInputRate = 23 * 1000,
                 maxInputEvents = 400,
                 maxInputSize = 2 * 1024 * 1024 * 1024,
                 maxInputFiles = 15,
                 maxLatency = 15 * 23,
                 periodicHarvestInterval = 20 * 60,
                 blockCloseDelay = 1200,
                 timePerEvent = 4, #I have to get some stats to set this properly
                 sizePerEvent = 1700, #I have to get some stats to set this properly
                 versionOverride = expressVersionOverride)

addExpressConfig(tier0Config, "HLTMonitor",
                 scenario = ppScenario,
                 data_tiers = [ "FEVTHLTALL" ],
                 write_dqm = True,
                 alca_producers = [],
                 dqm_sequences = [ "@HLTMon" ],
                 reco_version = defaultCMSSWVersion,
                 multicore = numberOfCores,
                 global_tag_connect = globalTagConnect,
                 global_tag = expressGlobalTag,
                 proc_ver = expressProcVersion,
                 maxInputRate = 23 * 1000,
                 maxInputEvents = 400,
                 maxInputSize = 2 * 1024 * 1024 * 1024,
                 maxInputFiles = 15,
                 maxLatency = 15 * 23,
                 periodicHarvestInterval = 20 * 60,
                 blockCloseDelay = 1200,
                 timePerEvent = 4, #I have to get some stats to set this properly
                 sizePerEvent = 1700, #I have to get some stats to set this properly
                 versionOverride = expressVersionOverride)

addExpressConfig(tier0Config, "Calibration",
                 scenario = alcaTestEnableScenario,
                 data_tiers = [ "RAW", "ALCARECO" ],
                 write_dqm = True,
                 alca_producers = [ "EcalTestPulsesRaw", "PromptCalibProdEcalPedestals" ],
                 reco_version = defaultCMSSWVersion,
                 multicore = numberOfCores,
                 global_tag_connect = globalTagConnect,
                 global_tag = expressGlobalTag,
                 proc_ver = expressProcVersion,
                 maxInputRate = 23 * 1000,
                 maxInputEvents = 100 * 1000 * 1000,
                 maxInputSize = 4 * 1024 * 1024 * 1024,
                 maxInputFiles = 10000,
                 maxLatency = 1 * 3600,
                 periodicHarvestInterval = 24 * 3600,
                 blockCloseDelay = 2 * 3600,
                 timePerEvent = 4,
                 sizePerEvent = 1700,
                 versionOverride = expressVersionOverride,
                 dataType = "data",
                 archivalNode = None,
                 tapeNode = None,
                 diskNode = None)

addExpressConfig(tier0Config, "ExpressAlignment",
                 scenario = alcaTrackingOnlyScenario,
                 data_tiers = [ "ALCARECO" ],
                 write_dqm = True,
                 alca_producers = [ "TkAlMinBias", "PromptCalibProdBeamSpotHP" ],
                 dqm_sequences = [ "DQMOfflineTracking" ],
                 reco_version = defaultCMSSWVersion,
                 multicore = numberOfCores,
                 global_tag_connect = globalTagConnect,
                 global_tag = expressGlobalTag,
                 proc_ver = expressProcVersion,
                 maxInputRate = 23 * 1000,
                 maxInputEvents = 100 * 1000 * 1000,
                 maxInputSize = 4 * 1024 * 1024 * 1024,
                 maxInputFiles = 10000,
                 maxLatency = 1 * 3600,
                 periodicHarvestInterval = 24 * 3600,
                 blockCloseDelay = 2 * 3600,
                 timePerEvent = 4,
                 sizePerEvent = 1700,
                 versionOverride = expressVersionOverride,
                 archivalNode = None,
                 tapeNode = None,
                 diskNode = None)

addExpressConfig(tier0Config, "ALCALUMIPIXELSEXPRESS",
                 scenario = alcaLumiPixelsScenario,
                 data_tiers = [ "ALCARECO" ],
                 write_dqm = True,
                 alca_producers = [ "AlCaPCCRandom", "PromptCalibProdLumiPCC" ],
                 dqm_sequences = [],
                 reco_version = defaultCMSSWVersion,
                 multicore = 1,
                 global_tag_connect = globalTagConnect,
                 global_tag = expressGlobalTag,
                 proc_ver = expressProcVersion,
                 maxInputRate = 23 * 1000,
                 maxInputEvents = 100 * 1000 * 1000,
                 maxInputSize = 4 * 1024 * 1024 * 1024,
                 maxInputFiles = 10000,
                 maxLatency = 1 * 3600,
                 periodicHarvestInterval = 24 * 3600,
                 blockCloseDelay = 2 * 3600,
                 timePerEvent = 4,
                 sizePerEvent = 1700,
                 versionOverride = expressVersionOverride,
                 archivalNode = None,
                 tapeNode = None,
                 diskNode = None)

#####################
### HI Tests 2018 ###
#####################

addExpressConfig(tier0Config, "HIExpress",
                 scenario = hiTestppScenario,
                 diskNode = "T2_CH_CERN",
                 data_tiers = [ "FEVT" ],
                 write_dqm = True,
                 alca_producers = [ "SiStripPCLHistos", "SiStripCalZeroBias", "SiStripCalMinBias", "SiStripCalMinBiasAAG",
                                    "TkAlMinBias", "SiPixelCalZeroBias", "DtCalib", 
                                    "PromptCalibProd", "PromptCalibProdSiStrip", "PromptCalibProdSiPixelAli",
                                    "PromptCalibProdSiStripGains", "PromptCalibProdSiStripGainsAAG", "PromptCalibProdSiPixel"
                                    ],
                 reco_version = defaultCMSSWVersion,
                 multicore = numberOfCores,
                 global_tag_connect = globalTagConnect,
                 global_tag = expressGlobalTag,
                 proc_ver = expressProcVersion,
                 maxInputRate = 23 * 1000,
                 maxInputEvents = 400,
                 maxInputSize = 2 * 1024 * 1024 * 1024,
                 maxInputFiles = 15,
                 maxLatency = 15 * 23,
                 periodicHarvestInterval = 20 * 60,
                 blockCloseDelay = 1200,
                 timePerEvent = 4,
                 sizePerEvent = 1700,
                 versionOverride = expressVersionOverride)

addExpressConfig(tier0Config, "HIExpressAlignment",
                 scenario = HIalcaTrackingOnlyScenario,
                 data_tiers = [ "ALCARECO", "RAW" ],
                 write_dqm = True,
                 alca_producers = [ "TkAlMinBias", "PromptCalibProdBeamSpotHP" ],
                 dqm_sequences = [ "DQMOfflineTracking" ],
                 reco_version = defaultCMSSWVersion,
                 raw_to_disk = True,
                 multicore = numberOfCores,
                 global_tag_connect = globalTagConnect,
                 global_tag = expressGlobalTag,
                 proc_ver = expressProcVersion,
                 maxInputRate = 23 * 1000,
                 maxInputEvents = 100 * 1000 * 1000,
                 maxInputSize = 4 * 1024 * 1024 * 1024,
                 maxInputFiles = 10000,
                 maxLatency = 1 * 3600,
                 periodicHarvestInterval = 24 * 3600,
                 blockCloseDelay = 2 * 3600,
                 timePerEvent = 4,
                 sizePerEvent = 1700,
                 versionOverride = expressVersionOverride,
                 diskNode = "T2_CH_CERN")

addExpressConfig(tier0Config, "HIHLTMonitor",
                 scenario = hiTestppScenario,
                 diskNode = "T2_CH_CERN",
                 data_tiers = [ "FEVTHLTALL" ],
                 write_dqm = True,
                 alca_producers = [],
                 dqm_sequences = [ "@HLTMon" ],
                 reco_version = defaultCMSSWVersion,
                 multicore = numberOfCores,
                 global_tag_connect = globalTagConnect,
                 global_tag = expressGlobalTag,
                 proc_ver = expressProcVersion,
                 maxInputRate = 23 * 1000,
                 maxInputEvents = 400,
                 maxInputSize = 2 * 1024 * 1024 * 1024,
                 maxInputFiles = 15,
                 maxLatency = 15 * 23,
                 periodicHarvestInterval = 20 * 60,
                 blockCloseDelay = 1200,
                 timePerEvent = 4, #I have to get some stats to set this properly
                 sizePerEvent = 1700, #I have to get some stats to set this properly
                 versionOverride = expressVersionOverride)

###################################
### Standard Physics PDs (2017) ###
###################################

datasets = [ "BTagCSV" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = True,
               write_dqm = True,
               dqm_sequences = [ "@common" ],
               physics_skims = [ "LogError", "LogErrorMonitor" ],
               scenario = ppScenario)

datasets = [ "BTagMu" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = True,
               write_dqm = True,
               dqm_sequences = [ "@common" ],
               physics_skims = [ "LogError", "LogErrorMonitor" ],
               scenario = ppScenario)

datasets = [ "Charmonium" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = True,
               write_reco = True,
               write_dqm = True,
               dqm_sequences = [ "@common" ],
               alca_producers = [ "TkAlJpsiMuMu" ],
               physics_skims = [ "BPHSkim", "MuonPOGJPsiSkim", "LogError", "LogErrorMonitor" ],
               scenario = ppScenario)

datasets = [ "Cosmics" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
           do_reco = True,
           write_reco = False,
           write_miniaod = False,
           write_dqm = True,
           alca_producers = [ "TkAlCosmics0T", "DtCalibCosmics" ],
           physics_skims = [ "CosmicSP", "CosmicTP", "LogError", "LogErrorMonitor" ],
           timePerEvent = 0.5,
           sizePerEvent = 155,
           scenario = cosmicsScenario)

datasets = [ "DisplacedJet" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = True,
               write_dqm = True,
               dqm_sequences = [ "@common" ],
               physics_skims = [ "EXODisplacedJet", "LogError", "LogErrorMonitor" ],
               scenario = ppScenario)

datasets = [ "DoubleEG" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = True,
               write_reco = True,
               write_dqm = True,
               alca_producers = [ "EcalUncalZElectron", "EcalUncalWElectron", "HcalCalIterativePhiSym", "HcalCalIsoTrkFilter" ],
               dqm_sequences = [ "@common", "@ecal", "@egamma" ],
               physics_skims = [ "ZElectron", "LogError", "LogErrorMonitor" ],
               scenario = ppScenario)

datasets = [ "DoubleMuon" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = True,
               write_reco = True,
               write_dqm = True,
               alca_producers = [ "TkAlZMuMu", "MuAlCalIsolatedMu", "MuAlOverlaps", "MuAlZMuMu", "DtCalib" ],
               dqm_sequences = [ "@common", "@muon", "@lumi", "@L1TMuon" ],
               physics_skims = [ "LogError", "LogErrorMonitor" ],
               scenario = ppScenario)

datasets = [ "DoubleMuonLowPU" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = True,
               write_reco = False,
               raw_to_disk = True,
               write_dqm = True,
               alca_producers = [ "TkAlZMuMu", "MuAlCalIsolatedMu", "MuAlOverlaps", "MuAlZMuMu", "DtCalib" ],
               dqm_sequences = [ "@common", "@muon", "@lumi", "@L1TMuon" ],
               physics_skims = [ "LogError", "LogErrorMonitor" ],
               timePerEvent = 1,
               scenario = ppScenario)

datasets = [ "DoubleMuonLowMass" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = True,
               write_dqm = True,
               dqm_sequences = [ "@common" ],
               physics_skims = [ "LogError", "LogErrorMonitor", "BPHSkim" ],
               scenario = ppScenario)

datasets = [ "EmptyBX" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = True,
               write_dqm = True,
               dqm_sequences = [ "@common" ],
               scenario = ppScenario)

datasets = [ "FSQJet1", "FSQJet2" ]

datasets += [ "FSQJets", "FSQJets1", "FSQJets2", "FSQJets3" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = True,
               write_dqm = True,
               dqm_sequences = [ "@common", "@jetmet" ],
               scenario = ppScenario)

datasets = [ "HINCaloJets", "HINPFJets" ]

datasets += [ "HINCaloJet40", "HINPFJet100", "HINCaloJet100", "HINCaloJetsOther", "HINPFJetsOther" ]

datasets += [ "HINMuon", "HINPhoton" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = True,
               write_dqm = True,
               dqm_sequences = [ "@common" ],
               scenario = ppScenario)

datasets = [ "HTMHT" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = True,
               write_dqm = True,
               dqm_sequences = [ "@common" ],
               physics_skims = [ "LogError", "LogErrorMonitor" ],
               timePerEvent = 9.4,
               sizePerEvent = 2000,
               scenario = ppScenario)

datasets = [ "HighMultiplicity" ]

datasets += [ "HighMultiplicityEOF0", "HighMultiplicityEOF1", "HighMultiplicityEOF2",
              "HighMultiplicityEOF3", "HighMultiplicityEOF4", "HighMultiplicityEOF5" ]

datasets += [ "HighMultiplicity85", "HighMultiplicity85EOF" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = True,
               write_reco = True,
               write_dqm = True,
               dqm_sequences = [ "@common" ],
               scenario = ppScenario)

# 05/07/2018 HighMultiplicityEOF needs to have 1sec per event
datasets = [ "HighMultiplicityEOF" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = True,
               write_reco = False,
               write_dqm = True,
               dqm_sequences = [ "@common" ],
               timePerEvent = 1,
               scenario = ppScenario)

datasets = [ "HighPtLowerPhotons", "HighPtPhoton30AndZ" ]

datasets += [ "ppForward", "HighPtLowerJets", "MuPlusX" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = True,
               write_dqm = True,
               dqm_sequences = [ "@common" ],
               scenario = ppScenario)

datasets = [ "JetHT" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = True,
               write_dqm = True,
               alca_producers = [ "HcalCalIsoTrkFilter", "HcalCalIsolatedBunchFilter" ],
               dqm_sequences = [ "@common", "@jetmet", "@hcal" ],
               physics_skims = [ "JetHTJetPlusHOFilter", "LogError", "LogErrorMonitor" ],
               timePerEvent = 5.7,
               sizePerEvent = 2250,
               scenario = ppScenario)

datasets = [ "MET" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = True,
               write_dqm = True,
               alca_producers = [ "HcalCalNoise" ],
               dqm_sequences = [ "@common", "@jetmet", "@hcal" ],
               physics_skims = [ "EXOMONOPOLE", "HighMET", "LogError", "LogErrorMonitor" ],
               scenario = ppScenario)

datasets = [ "MuOnia" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = True,
               write_dqm = True,
               alca_producers = [ "TkAlUpsilonMuMu" ],
               dqm_sequences = [ "@common", "@muon" ],
               physics_skims = [ "LogError", "LogErrorMonitor", "BPHSkim" ],
               scenario = ppScenario)

datasets = [ "MuonEG" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = True,
               write_dqm = True,
               dqm_sequences = [ "@common" ],
               physics_skims = [ "TopMuEG", "LogError", "LogErrorMonitor" ],
               scenario = ppScenario)

datasets = [ "NoBPTX" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = True,
               write_reco = True,
               write_dqm = True,
               alca_producers = [ "TkAlCosmicsInCollisions" ],
               dqm_sequences = [ "@common" ],
               physics_skims = [ "EXONoBPTXSkim", "LogError", "LogErrorMonitor" ],
               scenario = ppScenario)

datasets = [ "SingleElectron" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = True,
               write_dqm = True,
               alca_producers = [ "EcalUncalWElectron", "EcalUncalZElectron", "HcalCalIterativePhiSym", "EcalESAlign" ],
               dqm_sequences = [ "@common", "@ecal", "@egamma", "@L1TEgamma" ],
               physics_skims = [ "LogError", "LogErrorMonitor" ],
               scenario = ppScenario)

datasets = [ "SingleMuon", "SingleMuonTnP" ] # SingleMuonTnP only for 2017 ppRef run

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = True,
               write_dqm = True,
               alca_producers = [ "TkAlMuonIsolated", "HcalCalIterativePhiSym", "DtCalib", "MuAlCalIsolatedMu", "MuAlOverlaps", "MuAlZMuMu", "HcalCalHO", "HcalCalHBHEMuonFilter" ],
               dqm_sequences = [ "@common", "@muon", "@lumi", "@L1TMuon" ],
               physics_skims = [ "MuonPOGSkim", "MuTau", "ZMu", "LogError", "LogErrorMonitor" ],
               scenario = ppScenario)

datasets = [ "SinglePhoton" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = True,
               write_dqm = True,
               dqm_sequences = [ "@common", "@ecal", "@egamma" ],
               physics_skims = [ "EXOMONOPOLE", "LogError", "LogErrorMonitor" ],
               scenario = ppScenario)

datasets = [ "EGamma" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = True,
               write_dqm = True,
               alca_producers = [ "EcalUncalZElectron", "EcalUncalWElectron", "HcalCalIterativePhiSym", "HcalCalIsoTrkFilter", "EcalESAlign" ],
               dqm_sequences = [ "@common", "@ecal", "@egamma", "@L1TEgamma" ],
               physics_skims = [ "EXOMONOPOLE", "ZElectron", "LogError", "LogErrorMonitor" ],
               scenario = ppScenario)

datasets = [ "Tau" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = True,
               write_dqm = True,
               dqm_sequences = [ "@common" ],
               physics_skims = [ "LogError", "LogErrorMonitor" ],
               scenario = ppScenario)

#############################################
### Standard Commisioning PDs (2017)      ###
#############################################

datasets = [ "Commissioning" ]

datasets += [ "Commissioning1", "Commissioning2", "Commissioning3", "Commissioning4",
              "CommissioningMuons", "CommissioningEGamma", "CommissioningTaus", "CommissioningSingleJet", "CommissioningDoubleJet"]

for dataset in datasets:
    addDataset(tier0Config, dataset,
           do_reco = True,
           write_dqm = True,
           alca_producers = [ "TkAlMinBias", "SiStripCalMinBias", "HcalCalIsoTrk", "HcalCalIsolatedBunchSelector" ],
           dqm_sequences = [ "@common", "@hcal" ],
           physics_skims = [ "EcalActivity", "LogError", "LogErrorMonitor" ],
           timePerEvent = 12,
           sizePerEvent = 4000,
           scenario = ppScenario)

datasets = [ "HcalNZS" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
           do_reco = True,
           write_dqm = True,
           dqm_sequences = [ "@common", "@hcal" ],
           alca_producers = [ "HcalCalMinBias" ],
           physics_skims = [ "LogError", "LogErrorMonitor" ],
           timePerEvent = 4.2,
           sizePerEvent = 1900,
           scenario = hcalnzsScenario)

datasets = [ "TestEnablesEcalHcal", "TestEnablesEcalHcalDQM" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = True,
               alca_producers = [ "EcalTestPulsesRaw", "PromptCalibProdEcalPedestals" ],
               dqm_sequences = [ "@common" ],
               scenario = alcaTestEnableScenario)

datasets = [ "OnlineMonitor", "EcalLaser" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = False,
               scenario = ppScenario)

datasets = [ "CosmicsForEventDisplay" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = False,
               write_miniaod = False,
               scenario = cosmicsScenario)

datasets = [ "L1Accept", "L1Accepts" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = False,
               dqm_sequences = [ "@common" ],
               scenario = ppScenario)

###########################
### special AlcaRaw PDs ###
###########################

datasets = [ "AlCaLumiPixels" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
           do_reco = True,
           write_reco = False, write_aod = False, write_miniaod = False, write_dqm = True,
           reco_split = alcarawSplitting,
           proc_version = alcarawProcVersion,
           alca_producers = [ "AlCaPCCZeroBias" ],
           dqm_sequences = [ "@common" ],
           timePerEvent = 0.02,
           sizePerEvent = 38,
           scenario = alcaLumiPixelsScenario)

datasets = [ "AlCaLumiPixels0", "AlCaLumiPixels1", "AlCaLumiPixels2", "AlCaLumiPixels3",
             "AlCaLumiPixels4", "AlCaLumiPixels5", "AlCaLumiPixels6", "AlCaLumiPixels7",
             "AlCaLumiPixels8", "AlCaLumiPixels9", "AlCaLumiPixels10", "AlCaLumiPixels11",
             "AlCaLumiPixels12" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
           do_reco = True,
           write_reco = False, write_aod = False, write_miniaod = False, write_dqm = True,
           reco_split = alcarawSplitting,
           proc_version = alcarawProcVersion,
           alca_producers = [ "AlCaPCCZeroBias" ],
           dqm_sequences = [ "@common" ],
           timePerEvent = 0.02,
           sizePerEvent = 38,
           scenario = alcaLumiPixelsScenario)

datasets = [ "AlCaPhiSym" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = False,
               raw_to_disk = True,
               alca_producers = [ "EcalCalPhiSym" ],
               scenario = ppScenario)

datasets = [ "AlCaP0" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = False,
               raw_to_disk = True,
               alca_producers = [ "EcalCalPi0Calib", "EcalCalEtaCalib" ],
               scenario = ppScenario)

########################################################
### HLTPhysics PDs                                   ###
########################################################

datasets = [ "HLTPhysics" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = True,
               write_reco = True,
               write_dqm = True,
               dqm_sequences = [ "@common", "@ecal", "@jetmet", "@hcal", "@L1TEgamma" ],
               alca_producers = [ "TkAlMinBias" ],
               physics_skims = [ "LogError", "LogErrorMonitor" ],
               scenario = ppScenario)

datasets = [ "HLTPhysics0", "HLTPhysics1", "HLTPhysics2",
             "HLTPhysics3", "HLTPhysics4", "HLTPhysics5",
             "HLTPhysics6", "HLTPhysics7", "HLTPhysics8",
             "HLTPhysics9", "HLTPhysics10" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = True,
               write_reco = True,
               write_dqm = True,
               dqm_sequences = [ "@common", "@ecal", "@jetmet", "@hcal", "@L1TEgamma" ],
               alca_producers = [ "TkAlMinBias" ],
               physics_skims = [ "LogError", "LogErrorMonitor" ],
               scenario = ppScenario)

datasets = [ "HLTPhysicsBunchTrains", "HLTPhysicsIsolatedBunch" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = True,
               write_dqm = True,
               alca_producers = [ "SiStripCalMinBias", "TkAlMinBias", "HcalCalIsoTrkFilter" ],
               physics_skims = [ "LogError", "LogErrorMonitor" ],
               scenario = ppScenario)

datasets = [ "HLTPhysicspart0", "HLTPhysicspart1",
             "HLTPhysicspart2", "HLTPhysicspart3", "HLTPhysicspart4",
             "HLTPhysicspart5", "HLTPhysicspart6", "HLTPhysicspart7"  ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = False,
               scenario = ppScenario)

datasets = [ "EphemeralHLTPhysics1", "EphemeralHLTPhysics2", "EphemeralHLTPhysics3",
             "EphemeralHLTPhysics4", "EphemeralHLTPhysics5", "EphemeralHLTPhysics6",
             "EphemeralHLTPhysics7", "EphemeralHLTPhysics8"]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = True,
               scenario = ppScenario)

datasets = [ "HLTPhysicsCosmics", "HLTPhysicsCosmics1", "HLTPhysicsCosmics2",
            "HLTPhysicsCosmics3", "HLTPhysicsCosmics4", "HLTPhysicsCosmics5",
            "HLTPhysicsCosmics6", "HLTPhysicsCosmics7", "HLTPhysicsCosmics8" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = True,
               write_miniaod = False,
               dqm_sequences = [ "@common" ],
               scenario = cosmicsScenario)

########################################################
### MinimumBias PDs                                  ###
########################################################

datasets = [ "MinimumBias" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = True,
               write_dqm = True,
               dqm_sequences = [ "@common", "@commonSiStripZeroBias", "@ecal", "@hcal", "@muon", "@jetmet" ],
               alca_producers = [ "SiStripCalZeroBias", "SiStripCalMinBias", "TkAlMinBias" ],
               scenario = ppScenario)

datasets = [ "MinimumBias0", "MinimumBias1", "MinimumBias2", "MinimumBias3",
             "MinimumBias4", "MinimumBias5", "MinimumBias6", "MinimumBias7",
             "MinimumBias8", "MinimumBias9", "MinimumBias10", "MinimumBias11",
             "MinimumBias12", "MinimumBias13", "MinimumBias14", "MinimumBias15",
             "MinimumBias16", "MinimumBias17", "MinimumBias18", "MinimumBias19",
             "MinimumBias20"
             ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = True,
               write_dqm = True,
               dqm_sequences = [ "@common", "@commonSiStripZeroBias", "@ecal", "@hcal", "@muon", "@jetmet" ],
               timePerEvent = 1,
               alca_producers = [ "SiStripCalZeroBias", "SiStripCalMinBias", "TkAlMinBias" ],
               scenario = ppScenario)

datasets = [ "L1MinimumBias" ]

datasets += [ "L1MinimumBiasHF1", "L1MinimumBiasHF2", "L1MinimumBiasHF3", "L1MinimumBiasHF4",
              "L1MinimumBiasHF5", "L1MinimumBiasHF6", "L1MinimumBiasHF7", "L1MinimumBiasHF8" ]

datasets += [ "L1MinimumBias0", "L1MinimumBias1", "L1MinimumBias2", "L1MinimumBias3", "L1MinimumBias4",
              "L1MinimumBias5", "L1MinimumBias6", "L1MinimumBias7", "L1MinimumBias8", "L1MinimumBias9",
              "L1MinimumBias10" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = True,
               write_dqm = True,
               dqm_sequences = [ "@common" ],
               scenario = ppScenario)

########################################################
### ZeroBias PDs                                     ###
########################################################

datasets = [ "ZeroBias" ]

datasets += [ "ZeroBias0", "ZeroBias1", "ZeroBias2",
             "ZeroBias3", "ZeroBias4", "ZeroBias5", "ZeroBias6",
             "ZeroBias7", "ZeroBias8", "ZeroBias9", "ZeroBias10",
             "ZeroBias11", "ZeroBias12", "ZeroBias13", "ZeroBias14",
             "ZeroBias15", "ZeroBias16", "ZeroBias17", "ZeroBias18",
             "ZeroBias19", "ZeroBias20" ]

datasets += [ "ZeroBiasIsolatedBunches", "ZeroBiasIsolatedBunches0", "ZeroBiasIsolatedBunches1", "ZeroBiasIsolatedBunches2",
             "ZeroBiasIsolatedBunches3", "ZeroBiasIsolatedBunches4", "ZeroBiasIsolatedBunches5", "ZeroBiasIsolatedBunches6",
             "ZeroBiasIsolatedBunches7", "ZeroBiasIsolatedBunches8", "ZeroBiasIsolatedBunches9", "ZeroBiasIsolatedBunches10" ]

datasets += [ "ZeroBiasIsolatedBunch", "ZeroBiasAfterIsolatedBunch",
             "ZeroBiasIsolatedBunch0", "ZeroBiasIsolatedBunch1", "ZeroBiasIsolatedBunch2",
             "ZeroBiasIsolatedBunch3", "ZeroBiasIsolatedBunch4", "ZeroBiasIsolatedBunch5" ]

datasets += [ "ZeroBiasBunchTrains0", "ZeroBiasBunchTrains1", "ZeroBiasBunchTrains2",
             "ZeroBiasBunchTrains3", "ZeroBiasBunchTrains4", "ZeroBiasBunchTrains5" ]

datasets += [ "ZeroBiasFirstBunchAfterTrain", "ZeroBiasFirstBunchInTrain" ]

datasets += [ "ZeroBiasPixelHVScan0", "ZeroBiasPixelHVScan1", "ZeroBiasPixelHVScan2",
             "ZeroBiasPixelHVScan3", "ZeroBiasPixelHVScan4", "ZeroBiasPixelHVScan5",
             "ZeroBiasPixelHVScan6", "ZeroBiasPixelHVScan7" ]

datasets += [ "ZeroBias8b4e1", "ZeroBias8b4e2", "ZeroBias8b4e3",
             "ZeroBias8b4e4", "ZeroBias8b4e5", "ZeroBias8b4e6",
             "ZeroBias8b4e7", "ZeroBias8b4e8", "ZeroBias8b4e10",
             "ZeroBias8b4e9" ]

datasets += [ "ZeroBiasNominalTrains1", "ZeroBiasNominalTrains2", "ZeroBiasNominalTrains3",
             "ZeroBiasNominalTrains4", "ZeroBiasNominalTrains5", "ZeroBiasNominalTrains6",
             "ZeroBiasNominalTrains7", "ZeroBiasNominalTrains8", "ZeroBiasNominalTrains10",
             "ZeroBiasNominalTrains9" ]

datasets += [ "ZeroBiasPD01", "ZeroBiasPD02", "ZeroBiasPD03",
             "ZeroBiasPD04", "ZeroBiasPD05", "ZeroBiasPD06",
             "ZeroBiasPD07", "ZeroBiasPD08", "ZeroBiasPD09",
             "ZeroBiasPD10"]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = True,
               write_reco = True,
               write_dqm = True,
               dqm_sequences = [ "@commonSiStripZeroBias", "@ecal", "@hcal", "@muon", "@jetmet" ],
               alca_producers = [ "SiStripCalZeroBias", "TkAlMinBias", "LumiPixelsMinBias", "SiStripCalMinBias", "AlCaPCCZeroBiasFromRECO" ],
               physics_skims = [ "LogError", "LogErrorMonitor" ],
               timePerEvent = 3.5,
               sizePerEvent = 1500,
               scenario = ppScenario)

datasets = [ "EphemeralZeroBias1", "EphemeralZeroBias2", "EphemeralZeroBias3",
             "EphemeralZeroBias4", "EphemeralZeroBias5", "EphemeralZeroBias6",
             "EphemeralZeroBias7", "EphemeralZeroBias8"]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = True,
               scenario = ppScenario)

########################################################
### Parking and Scouting PDs                         ###
########################################################

datasets = [ "ParkingMonitor" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = True,
               dqm_sequences = [ "@common" ],
               write_reco = False, write_aod = False, write_miniaod = True, write_dqm = True,
               scenario = ppScenario)

datasets = [ "ParkingScoutingMonitor" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = True,
               dqm_sequences = [ "@common" ],
               write_reco = False, write_aod = False, write_miniaod = True, write_dqm = True,
               scenario = ppScenario)

datasets = [ "ParkingMuon", "ParkingHT" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = False,
               scenario = ppScenario)

datasets = [ "ParkingBPH1", "ParkingBPH2", "ParkingBPH3", "ParkingBPH4", "ParkingBPH5", "ParkingBPH6" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = False,
               tape_node = None,
               disk_node = None,
               scenario = ppScenario)

# Parking PD to be PR'ed at CSCS
datasets = [ "ParkingBPHPromptCSCS" ]
 
for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = False,
               siteWhitelist = [ "T0_CH_CSCS_HPC" ],
               scenario = ppScenario)
 
datasets = [ "RPCMonitor" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = False,
               scenario = ppScenario)

datasets = [ "ScoutingMonitor" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = True,
               dqm_sequences = [ "@common" ],
               write_reco = False, write_aod = False, write_miniaod = True, write_dqm = True,
               scenario = ppScenario)

datasets = [ "ScoutingCaloCommissioning", "ScoutingCaloHT", "ScoutingCaloMuon",
             "ScoutingPFCommissioning", "ScoutingPFHT", "ScoutingPFMuon" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = False,
               scenario = ppScenario)

datasets = [ "ParkingHT410to430", "ParkingHT500to550", "ParkingHT430to450", "ParkingHT470to500", "ParkingHT450to470" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = False,
               scenario = ppScenario)

datasets = [ "ParkingHT550to650", "ParkingHT650" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = False,
               scenario = ppScenario)

datasets = [ "ParkingHLTPhysics", "ParkingHLTPhysics0", "ParkingHLTPhysics1",
             "ParkingHLTPhysics2", "ParkingHLTPhysics3", "ParkingHLTPhysics4",
             "ParkingHLTPhysics5", "ParkingHLTPhysics6", "ParkingHLTPhysics7",
             "ParkingHLTPhysics8", "ParkingHLTPhysics9", "ParkingHLTPhysics10",
             "ParkingHLTPhysics11", "ParkingHLTPhysics12", "ParkingHLTPhysics13",
             "ParkingHLTPhysics14", "ParkingHLTPhysics15", "ParkingHLTPhysics16",
             "ParkingHLTPhysics17", "ParkingHLTPhysics18", "ParkingHLTPhysics19",
             "ParkingHLTPhysics20" ]

datasets += [ "ParkingZeroBias", "ParkingZeroBias0",
             "ParkingZeroBias1", "ParkingZeroBias2", "ParkingZeroBias3",
             "ParkingZeroBias4", "ParkingZeroBias5", "ParkingZeroBias6",
             "ParkingZeroBias7", "ParkingZeroBias8", "ParkingZeroBias9",
             "ParkingZeroBias10", "ParkingZeroBias11", "ParkingZeroBias12",
             "ParkingZeroBias13", "ParkingZeroBias14", "ParkingZeroBias15",
             "ParkingZeroBias16", "ParkingZeroBias17", "ParkingZeroBias18",
             "ParkingZeroBias19", "ParkingZeroBias20" ]

datasets += [ "AlCaElectron", "VRRandom", "VRRandom0", "VRRandom1", "VRRandom2", "VRRandom3",
             "VRRandom4", "VRRandom5", "VRRandom6", "VRRandom7", "VRZeroBias", "VirginRaw" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = False,
               scenario = ppScenario)

#########################################
### New PDs for pp Reference Run 2017 ###
#########################################

datasets = [ "HighEGJet", "LowEGJet" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = True,
               write_dqm = True,
               dqm_sequences = [ "@common", "@ecal", "@egamma", "@hcal", "@jetmet" ],
               scenario = ppScenario)

datasets = [ "HeavyFlavor", "SingleTrack" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = True,
               write_dqm = True,
               dqm_sequences = [ "@common" ],
               scenario = ppScenario)

datasets = [ "HIZeroBias1", "HIZeroBias2", "HIZeroBias3", "HIZeroBias4",
             "HIZeroBias5", "HIZeroBias6", "HIZeroBias7", "HIZeroBias8",
             "HIZeroBias9", "HIZeroBias10", "HIZeroBias11", "HIZeroBias12" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = True,
               write_dqm = True,
               dqm_sequences = [ "@commonSiStripZeroBias", "@ecal", "@hcal", "@muon" ],
               alca_producers = [ "SiStripCalZeroBias", "TkAlMinBias", "LumiPixelsMinBias", "SiStripCalMinBias", "AlCaPCCZeroBiasFromRECO" ],
               timePerEvent = 3.5,
               sizePerEvent = 1500,
               scenario = ppScenario)

datasets = [ "Totem12", "Totem34" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = True,
               dqm_sequences = [ "@common" ],
               scenario = ppScenario)

################################
### Low PU collisions 13 TeV ###
################################

datasets = [ "CastorJets", "EGMLowPU", "FullTrack" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = True,
               write_dqm = True,
               dqm_sequences = [ "@common" ],
               scenario = ppScenario)

datasets = [ "HcalHPDNoise" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = True,
               dqm_sequences = [ "@common" ],
               physics_skims = [ "LogError", "LogErrorMonitor" ],
               scenario = ppScenario)

datasets = [ "HINMuon_HFveto" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = True,
               dqm_sequences = [ "@common" ],
               scenario = ppScenario)

################################
### Special Totem runs       ###
################################

datasets = [ "TOTEM_minBias", "TOTEM_romanPots", "ToTOTEM", "ToTOTEM_DoubleJet32_0", "ToTOTEM_DoubleJet32_1",
             "ToTOTEM_DoubleJet32_2", "ToTOTEM_DoubleJet32_3", "TOTEM_zeroBias", "ZeroBiasTotem", "MinimumBiasTotem",
             "TOTEM_minBias1", "TOTEM_minBias2", "TOTEM_romanPots1", "TOTEM_romanPots2", "TOTEM_romanPots2_0",
             "TOTEM_romanPots2_1", "TOTEM_romanPots2_2", "TOTEM_romanPots2_3", "TOTEM_romanPots2_4",
             "TOTEM_romanPots2_5", "TOTEM_romanPots2_6", "TOTEM_romanPots2_7", "TOTEM_romanPots3",
             "TOTEM_romanPotsTTBB_0", "TOTEM_romanPotsTTBB_1", "TOTEM_romanPotsTTBB_2", "TOTEM_romanPotsTTBB_3",
             "TOTEM_romanPotsTTBB_4", "TOTEM_romanPotsTTBB_5", "TOTEM_romanPotsTTBB_6", "TOTEM_romanPotsTTBB_7" ]

datasets += [ "Totem1", "Totem2", "Totem3", "Totem4" ]

### TOTEM datasets for 90m and LowPileUp menu - 2018/06/22
datasets += [ "HFvetoTOTEM", "JetsTOTEM" ]

datasets += [ "RandomTOTEM1", "RandomTOTEM2", "RandomTOTEM3", "RandomTOTEM4" ]

datasets += [ "TOTEM10", "TOTEM11", "TOTEM12", "TOTEM13", "TOTEM20", "TOTEM21", "TOTEM22",
              "TOTEM23", "TOTEM3", "TOTEM40", "TOTEM41", "TOTEM42", "TOTEM43" ]

datasets += [ "ZeroBiasTOTEM1", "ZeroBiasTOTEM2", "ZeroBiasTOTEM3", "ZeroBiasTOTEM4" ]
### TOTEM datasets for 90m and LowPileUp menu - 2018/06/22

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = True,
               write_reco = True,
               dqm_sequences = [ "@common" ],
               scenario = ppScenario)

### TOTEM EGamma dataset for 90m and LowPileUp with egamma dqm sequence
datasets = [ "MuonEGammaTOTEM" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = True,
               write_reco = False,
               write_dqm = True,
               dqm_sequences = [ "@common", "@egamma" ],
               scenario = ppScenario)

################################
### 50 ns Physics Menu       ###
################################

datasets = [ "L1TechBPTXPlusOnly", "L1TechBPTXMinusOnly", "L1TechBPTXQuiet" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = True,
               dqm_sequences = [ "@common" ],
               scenario = ppScenario)

datasets = [ "SingleMu" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = True,
               dqm_sequences = [ "@common" ],
               scenario = ppScenario)

datasets = [ "DoubleMu" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = True,
               dqm_sequences = [ "@common" ],
               alca_producers = [ "TkAlZMuMu", "TkAlJpsiMuMu", "TkAlUpsilonMuMu", "MuAlCalIsolatedMu", "MuAlOverlaps", "MuAlZMuMu", "DtCalib", "HcalCalIsoTrkFilter" ],
               physics_skims = [ "Onia" ],
               scenario = ppScenario)

datasets = [ "DoublePhoton" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = True,
               write_dqm = True,
               dqm_sequences = [ "@common", "@ecal", "@egamma" ],
               scenario = ppScenario)


#########################################
### New PDs for pp Reference Run 2015 ###
#########################################

# new PD with same name added for 2017 ppRef
# keeping this config for reference
# addDataset(tier0Config, "HeavyFlavor",
#           do_reco = True,
#            write_dqm = True,
#            dqm_sequences = [ "@common" ],
#            physics_skims = [ "D0Meson" ],
#            scenario = ppScenario)

addDataset(tier0Config, "HighPtJet80",
           do_reco = True,
           write_dqm = True,
           dqm_sequences = [ "@common" ],
           physics_skims = [ "HighPtJet" ],
           scenario = ppScenario)

addDataset(tier0Config, "SingleMuHighPt",
           do_reco = True,
           write_dqm = True,
           alca_producers = [ "TkAlMuonIsolated", "HcalCalIterativePhiSym", "DtCalib", "MuAlCalIsolatedMu", "MuAlOverlaps", "MuAlZMuMu" ],
           dqm_sequences = [ "@common" ],
           physics_skims = [ "ZMM" ],
           scenario = ppScenario)

addDataset(tier0Config, "SingleMuLowPt",
           do_reco = True,
           write_dqm = True,
           alca_producers = [ "TkAlMuonIsolated", "HcalCalIterativePhiSym", "DtCalib", "MuAlCalIsolatedMu", "MuAlOverlaps", "MuAlZMuMu" ],
           dqm_sequences = [ "@common" ],
           scenario = ppScenario)

###############################
### ExpressPA configuration ###
###############################

addExpressConfig(tier0Config, "ExpressPA",
                 scenario = hiScenario,
                 data_tiers = [ "FEVT" ],
                 write_dqm = True,
                 alca_producers = [ "SiStripCalZeroBias", "TkAlMinBias", "DtCalib", "SiStripCalMinBias",
                                    "SiStripCalMinBiasAfterAbortGap", "LumiPixelsMinBias", "PromptCalibProd",
                                    "PromptCalibProdSiStrip", "PromptCalibProdSiPixelAli", "PromptCalibProdSiStripGains",
                                    "PromptCalibProdSiStripGainsAfterAbortGap", "SiStripPCLHistos" ],
                 reco_version = defaultCMSSWVersion,
                 multicore = numberOfCores,
                 global_tag_connect = globalTagConnect,
                 global_tag = expressGlobalTag,
                 proc_ver = expressProcVersion,
                 maxInputRate = 23 * 1000,
                 maxInputEvents = 400,
                 maxInputSize = 2 * 1024 * 1024 * 1024,
                 maxInputFiles = 15,
                 maxLatency = 15 * 23,
                 periodicHarvestInterval = 20 * 60,
                 blockCloseDelay = 1200,
                 timePerEvent = 4,
                 sizePerEvent = 1700,
                 versionOverride = expressVersionOverride)

addExpressConfig(tier0Config, "HLTMonitorPA",
                 scenario = hiScenario,
                 data_tiers = [ "FEVTHLTALL" ],
                 write_dqm = True,
                 alca_producers = [],
                 dqm_sequences = [ "@HLTMonPA" ],
                 reco_version = defaultCMSSWVersion,
                 multicore = numberOfCores,
                 global_tag_connect = globalTagConnect,
                 global_tag = expressGlobalTag,
                 proc_ver = expressProcVersion,
                 maxInputRate = 23 * 1000,
                 maxInputEvents = 400,
                 maxInputSize = 2 * 1024 * 1024 * 1024,
                 maxInputFiles = 15,
                 maxLatency = 15 * 23,
                 periodicHarvestInterval = 20 * 60,
                 blockCloseDelay = 1200,
                 timePerEvent = 4, #I have to get some stats to set this properly
                 sizePerEvent = 1700, #I have to get some stats to set this properly
                 versionOverride = expressVersionOverride)

#########################################
### New PDs for PARun 2016 ###
#########################################

datasets = [ "PAHighMultiplicity0", "PAHighMultiplicity1", "PAHighMultiplicity2", "PAHighMultiplicity3",
             "PAHighMultiplicity4", "PAHighMultiplicity5", "PAHighMultiplicity6", "PAHighMultiplicity7" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = True,
               write_dqm = True,
               dqm_sequences = [ "@common" ],
               scenario = hiScenario)

addDataset(tier0Config, "PACastor",
           do_reco = True,
           write_dqm = True,
           dqm_sequences = [ "@common" ],
           scenario = hiScenario)

addDataset(tier0Config, "PAForward",
           do_reco = True,
           write_dqm = True,
           dqm_sequences = [ "@common" ],
           scenario = hiScenario)

addDataset(tier0Config, "PADoubleMuon",
           do_reco = True,
           write_dqm = True,
           dqm_sequences = [ "@common", "@muon" ],
           alca_producers = [ "TkAlMuonIsolatedPA", "TkAlZMuMuPA", "TkAlUpsilonMuMuPA", "DtCalib" ],
           scenario = hiScenario)

addDataset(tier0Config, "PASingleMuon",
           do_reco = True,
           write_dqm = True,
           dqm_sequences = [ "@common", "@muon" ],
           alca_producers = [ "TkAlMuonIsolatedPA", "DtCalib" ],
           physics_skims = [ "PAZMM" ],
           scenario = hiScenario)

datasets = [ "PADTrack1", "PADTrack2" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = True,
               write_dqm = True,
               dqm_sequences = [ "@common" ],
               scenario = hiScenario)

addDataset(tier0Config, "PAEGJet1",
           do_reco = True,
           write_dqm = True,
           dqm_sequences = [ "@common", "@ecal", "@egamma", "@hcal", "@jetmet" ],
           physics_skims = [ "PAZEE" ],
           scenario = hiScenario)

datasets = [ "PAMinimumBias1", "PAMinimumBias2", "PAMinimumBias3", "PAMinimumBias4",
             "PAMinimumBias5", "PAMinimumBias6", "PAMinimumBias7", "PAMinimumBias8",
             "PAMinimumBias9", "PAMinimumBias10", "PAMinimumBias11", "PAMinimumBias12",
             "PAMinimumBias13", "PAMinimumBias14", "PAMinimumBias15", "PAMinimumBias16",
             "PAMinimumBias17", "PAMinimumBias18", "PAMinimumBias19", "PAMinimumBias20" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = True,
               write_dqm = True,
               dqm_sequences = [ "@common" ],
               alca_producers = [ "SiStripCalMinBias", "TkAlMinBias" ],
               scenario = hiScenario)

addDataset(tier0Config, "PAMinimumBiasBkg",
           do_reco = True,
           write_dqm = True,
           dqm_sequences = [ "@common" ],
           physics_skims = [ "PAMinBias" ],
           scenario = hiScenario)


addDataset(tier0Config, "PAEmptyBX",
           do_reco = True,
           write_dqm = True,
           dqm_sequences = [ "@common" ],
           scenario = hiScenario)

#############################
###   PDs pA VdM scan     ###
#############################

datasets = [ "PAL1AlwaysTrue0", "PAL1AlwaysTrue1", "PAL1AlwaysTrue2", "PAL1AlwaysTrue3",
             "PAL1AlwaysTrue4", "PAL1AlwaysTrue5", "PAL1AlwaysTrue6", "PAL1AlwaysTrue7",
             "PAL1AlwaysTrue8", "PAL1AlwaysTrue9" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = True,
               write_dqm = True,
               alca_producers = [ "LumiPixelsMinBias" ],
               dqm_sequences = [ "@common" ],
               scenario = hiScenario)

datasets = [ "PAMinimumBiasHFOR0", "PAMinimumBiasHFOR1", "PAMinimumBiasHFOR2" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = True,
               write_dqm = True,
               alca_producers = [ "LumiPixelsMinBias" ],
               dqm_sequences = [ "@common" ],
               scenario = hiScenario)

datasets = [ "PAZeroBias0", "PAZeroBias1", "PAZeroBias2", "PAZeroBias3",
             "PAZeroBias4", "PAZeroBias5", "PAZeroBias6", "PAZeroBias7",
             "PAZeroBias8", "PAZeroBias9" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = True,
               write_dqm = True,
               alca_producers = [ "LumiPixelsMinBias" ],
               dqm_sequences = [ "@commonSiStripZeroBias", "@ecal", "@hcal", "@muon" ],
               scenario = hiScenario)

addDataset(tier0Config, "PADoubleMuOpen",
           do_reco = True,
           write_dqm = True,
           alca_producers = [ "LumiPixelsMinBias" ],
           dqm_sequences = [ "@common", "@muon" ],
           scenario = hiScenario)

#####################
### HI TESTS 2018 ###
#####################

datasets = [ "HITestFull", "HITestReduced" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = True,
               raw_to_disk = True,
               write_dqm = True,
               dqm_sequences = [ "@common" ],
               scenario = hiTestppScenario)

datasets = [ "HIHardProbesPrescaled", "HIHardProbesPeripheral", "HICommissioning",
             "HICastor" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               write_miniaod = False,
               do_reco = True,
               raw_to_disk = True,
               write_dqm = True,
               dqm_sequences = [ "@common" ],
               scenario = hiTestppScenario)

datasets = [ "HIHeavyFlavor", "HIHighMultiplicityETTAsym" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               write_miniaod = False,
               do_reco = True,
               write_dqm = True,
               reco_split = hiRecoSplitting,
               dqm_sequences = [ "@common" ],
               scenario = hiTestppScenario)

datasets = [ "HIMinimumBiasReducedFormat0", "HILowMultiplicityReducedFormat", "HILowMultiplicity" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               write_miniaod = False,
               do_reco = True,
               raw_to_disk = True,
               write_dqm = True,
               dqm_sequences = [ "@common" ],
               scenario = hiTestppScenario)

# CMS VdM scan PDs
datasets = [ "HICentralityVetoReducedFormat0", "HICentralityVetoReducedFormat1", "HICentralityVetoReducedFormat2",
             "HICentralityVetoReducedFormat3", "HICentralityVetoReducedFormat4", "HICentralityVetoReducedFormat5",
             "HICentralityVetoReducedFormat6", "HICentralityVetoReducedFormat7", "HICentralityVetoReducedFormat8",
             "HICentralityVetoReducedFormat9", "HICentralityVetoReducedFormat10", "HICentralityVetoReducedFormat11" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               write_miniaod = False,
               do_reco = True,
               write_dqm = True,
               dqm_sequences = [ "@common" ],
               scenario = hiTestppScenario)

datasets = [ "HIMinimumBiasReducedFormat1", "HIMinimumBiasReducedFormat10", "HIMinimumBiasReducedFormat11", 
             "HIMinimumBiasReducedFormat2", "HIMinimumBiasReducedFormat3", "HIMinimumBiasReducedFormat4", 
             "HIMinimumBiasReducedFormat5", "HIMinimumBiasReducedFormat6", "HIMinimumBiasReducedFormat7", 
             "HIMinimumBiasReducedFormat8", "HIMinimumBiasReducedFormat9" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               write_miniaod = False,
               do_reco = True,
               raw_to_disk = True,
               write_dqm = False,
               dqm_sequences = [ "@none" ],
               scenario = hiTestppScenario)

datasets = [ "HIForward" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               write_miniaod = False,
               do_reco = True,
               raw_to_disk = True,
               write_dqm = True,
               dqm_sequences = [ "@commonSiStripZeroBias" ],
               scenario = hiTestppScenario)

datasets = [ "HIMinimumBias0", "HIMinimumBias1" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               write_miniaod = False,
               do_reco = True,
               write_dqm = True,
               dqm_sequences = [ "@commonSiStripZeroBias", "@hcal" ],
               scenario = hiTestppScenario)

datasets = [ "HIMinimumBias2",
             "HIMinimumBias3", "HIMinimumBias4", "HIMinimumBias5",
             "HIMinimumBias6", "HIMinimumBias7", "HIMinimumBias8",
             "HIMinimumBias9", "HIMinimumBias10", "HIMinimumBias11",
             "HIMinimumBias12", "HIMinimumBias13", "HIMinimumBias14",
             "HIMinimumBias15", "HIMinimumBias16", "HIMinimumBias17",
             "HIMinimumBias18","HIMinimumBias19" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               write_miniaod = False,
               do_reco = True,
               raw_to_disk = True,
               write_dqm = False,
               dqm_sequences = [ "@none" ],
               scenario = hiTestppScenario)

datasets = [ "HIHcalNZS" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               write_miniaod = False,
               do_reco = True,
               raw_to_disk = True,
               write_dqm = True,
               alca_producers = [ "HcalCalMinBias" ],
               dqm_sequences = [ "@common", "@hcal" ],
               scenario = hcalnzsScenario)

datasets = [ "HIHLTPhysics" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               write_miniaod = False,
               do_reco = True,
               raw_to_disk = True,
               write_dqm = True,
               alca_producers = [ "TkAlMinBias" ],
               dqm_sequences = [ "@common" ],
               scenario = hiTestppScenario)

datasets = [ "HIHardProbesLower" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               write_miniaod = False,
               do_reco = True,
               raw_to_disk = True,
               write_dqm = True,
               reco_split = hiRecoSplitting,
               dqm_sequences = [ "@common", "@ecal", "@egamma" ],
               scenario = hiTestppScenario)

datasets = [ "HIHardProbes" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               write_miniaod = False,
               do_reco = True,
               raw_to_disk = True,
               write_dqm = True,
               reco_split = hiRecoSplitting,
               alca_producers = [ "TkAlMinBias", "HcalCalIterativePhiSym", "SiStripCalSmallBiasScan" ],
               dqm_sequences = [ "@common", "@ecal", "@hcal", "@jetmet", "@egamma" ],
               physics_skims = [ "PbPbEMu", "PbPbZEE" ],
               scenario = hiTestppScenario)

datasets = [ "HISingleMuon" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = True,
               write_dqm = True,
               raw_to_disk = True,
               write_miniaod = False,
               reco_split = hiRecoSplitting,
               alca_producers = [ "TkAlZMuMu", "TkAlMuonIsolated", "DtCalib", "HcalCalIterativePhiSym" ],
               dqm_sequences = [ "@common", "@muon", "@lumi" ],
               physics_skims = [ "PbPbZMu" ],
               scenario = hiTestppScenario)

datasets = [ "HIDoubleMuon", "HIDoubleMuonPsiPeri" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = True,
               write_reco = False,
               raw_to_disk = True,
               write_miniaod = False,
               write_dqm = True,
               reco_split = hiRecoSplitting,
               alca_producers = [ "TkAlJpsiMuMu", "TkAlUpsilonMuMu" ],
               dqm_sequences = [ "@common", "@muon", "@lumi" ],
               physics_skims = [ "PbPbZMM" ],
               scenario = hiTestppScenario)

datasets = [ "HIOnlineMonitor", "HITrackerNZS" ]

for dataset in datasets:
    addDataset(tier0Config, dataset,
               do_reco = False,
               raw_to_disk = True,
               scenario = hiTestppScenario)

#######################
### ignored streams ###
#######################

ignoreStream(tier0Config, "Error")
ignoreStream(tier0Config, "HLTMON")
ignoreStream(tier0Config, "EventDisplay")
ignoreStream(tier0Config, "DQM")
ignoreStream(tier0Config, "DQMEventDisplay")
ignoreStream(tier0Config, "LookArea")
ignoreStream(tier0Config, "DQMOffline")
ignoreStream(tier0Config, "streamHLTRates")
ignoreStream(tier0Config, "streamL1Rates")
ignoreStream(tier0Config, "streamDQMRates")

###################################
### currently inactive settings ###
###################################

##ignoreStream(tier0Config, "Express")
##addRegistrationConfig(tier0Config, "UserStreamExample1",
##                      primds = "ExamplePrimDS1",
##                      acq_era = "AcqEra1",
##                      proc_string = "OptionalProcString",
##                      proc_version = "v1",
##                      data_tier = "RAW")
##
##addConversionConfig(tier0Config, "UserStreamExample",
##                    primds = "PrimDSTest6",
##                    acq_era = "AquEraTest6",
##                    proc_string = "ProcStringTest6",
##                    proc_version = "v6",
##                    data_tier = "RAW",
##                    conv_type = "streamer")

if __name__ == '__main__':
    print(tier0Config)



"""
_StorageManagerAPI_

Contains all the code for interfacing with the StorageManager

"""
import logging
import threading
import time

from WMCore.DAOFactory import DAOFactory

knownStreamers = set()

def injectNewData(dbInterfaceStorageManager,
                  dbInterfaceHltConf,
                  dbInterfaceSMNotify,
                  streamerPNN,
                  minRun = None,
                  maxRun = None,
                  injectRun = None):
    """
    _injectNewData_

    Replaces the old-style file notification injecton into the Tier0.

    Queries the StorageManager database for new data and injects it into the Tier0.

    These queries will find duplicates, ie. data that was already found and
    processed in a previous polling cycle. Code has to be robust against that.

    Needs to be passed the PNN on which streamer files are located

    """
    logging.debug("injectNewData()")
    myThread = threading.currentThread()

    daoFactory = DAOFactory(package = "T0.WMBS",
                            logger = logging,
                            dbinterface = myThread.dbi)

    daoFactoryStorageManager = DAOFactory(package = "T0.WMBS",
                                          logger = logging,
                                          dbinterface = dbInterfaceStorageManager)

    daoFactoryHltConf = DAOFactory(package = "T0.WMBS",
                                   logger = logging,
                                   dbinterface = dbInterfaceHltConf)

    if dbInterfaceSMNotify:
        daoFactorySMNotify = DAOFactory(package = "T0.WMBS",
                                        logger = logging,
                                        dbinterface = dbInterfaceSMNotify)
        insertFileStatusDAO = daoFactorySMNotify(classname = "SMNotification.InsertOfflineFileStatus")

    getNewDataDAO = daoFactoryStorageManager(classname = "StorageManager.GetNewData")
    getRunInfoDAO = daoFactoryHltConf(classname = "StorageManager.GetRunInfo")
    insertRunDAO = daoFactory(classname = "RunConfig.InsertRun")
    insertStreamDAO = daoFactory(classname = "RunConfig.InsertStream")
    insertCMSSWVersionDAO = daoFactory(classname = "RunConfig.InsertCMSSWVersion")
    insertStreamCMSSWVersionDAO = daoFactory(classname = "RunConfig.InsertStreamCMSSWVersion")
    insertLumiDAO = daoFactory(classname = "RunConfig.InsertLumiSection")
    insertStreamerDAO = daoFactory(classname = "RunConfig.InsertStreamer")

    newData = getNewDataDAO.execute(minRun = minRun,
                                    maxRun = maxRun,
                                    injectRun = injectRun, 
                                    transaction = False)

    # remove already processed files
    newData[:] = [newFile for newFile in newData if newFile['p5_id'] not in knownStreamers]

    logging.debug("StoragemanagerAPI: found %d new files", len(newData))

    newRuns = set()
    newRunStreams = {}
    for newFile in newData:

        run = newFile['run']
        stream = newFile['stream']

        newRuns.add(newFile['run'])

        if run not in newRunStreams:
            newRunStreams[run] = set()
        if stream not in newRunStreams[run]:
            newRunStreams[run].add(stream)

    logging.debug("StoragemanagerAPI: found %d new runs", len(newRuns))

    cmsswVersions = set()
    streams = set()
    bindRunHltKey = []
    bindRunStreamCMSSW = []
    for run in sorted(list(newRuns)):
        (hltkey, cmssw) = getRunInfoDAO.execute(run = run, transaction = False)
        logging.debug("StorageManagerAPI: run = %d, hltkey = %s, cmssw = %s", run, hltkey, cmssw)
        if hltkey and cmssw:
            cmssw = '_'.join(cmssw.split('_')[0:4]) # only consider base release
            cmsswVersions.add(cmssw)
            bindRunHltKey.append( { 'RUN': run,
                                    'HLTKEY': hltkey } )
            for stream in newRunStreams[run]:
                streams.add(stream)
                bindRunStreamCMSSW.append( { 'RUN': run,
                                             'STREAM': stream,
                                             'VERSION': cmssw } )
        else:
            # can't retrieve hltkey and cmssw for run, ignore any data for it
            newRuns.remove(run)

    if len(bindRunHltKey) > 0:
        insertRunDAO.execute(binds = bindRunHltKey, transaction = False)

    bindStream = []
    for stream in streams:
        bindStream.append( { 'STREAM': stream } )
    if len(bindStream) > 0:
        insertStreamDAO.execute(binds = bindStream, transaction = False)

    bindCMSSW = []
    for cmssw in cmsswVersions:
        bindCMSSW.append( { 'VERSION': cmssw } )
    if len(bindCMSSW) > 0:
        insertCMSSWVersionDAO.execute(binds = bindCMSSW, transaction = False)

    if len(bindRunStreamCMSSW) > 0:
        insertStreamCMSSWVersionDAO.execute(binds = bindRunStreamCMSSW, transaction = False)

    lumis = set()
    bindStreamer = []
    bindInsertFileStatus = []
    for newFile in newData:

        run = newFile['run']

        if run not in newRuns:
            continue

        lumi = newFile['lumi']
        lumis.add((run,lumi))

        if newFile['filename'] == 'run289461_ls0020_streamExpressCosmics_StorageManager.dat':
            newFile['path'] = '/store/t0streamer/Data/ExpressCosmics/000/289/461'

        bindStreamer.append( { 'LFN': newFile['path'] + '/' + newFile['filename'],
                               'P5_ID': newFile['p5_id'],
                               'RUN': run,
                               'LUMI': lumi,
                               'STREAM': newFile['stream'],
                               'FILESIZE': newFile['filesize'],
                               'EVENTS': newFile['events'],
                               'TIME': int(time.time()) } )

        if dbInterfaceSMNotify:
            bindInsertFileStatus.append( { 'P5_ID': newFile['p5_id'],
                                           'FILENAME': newFile['filename'] } )

    bindLumi = []
    for lumi in lumis:
        bindLumi.append( { 'RUN': lumi[0],
                           'LUMI': lumi[1] } )
    if len(bindLumi) > 0:
        insertLumiDAO.execute(binds = bindLumi, transaction = False)

    if len(bindStreamer) > 0:
        insertStreamerDAO.execute(streamerPNN, binds = bindStreamer, transaction = False)

    if len(bindInsertFileStatus) > 0:
        insertFileStatusDAO.execute(bindInsertFileStatus, transaction = False)

    for x in bindStreamer:
        knownStreamers.add(x['P5_ID'])

    return

def markRepacked(dbInterfaceSMNotify):
    """
    _markRepacked_

    Find all finished streamers for closed all run/stream
    Update the StorageManager notification table
    Update the streamer status to finished (deleted = 1)

    """
    if not dbInterfaceSMNotify:
        return

    logging.debug("updateFileStatus()")
    myThread = threading.currentThread()

    daoFactory = DAOFactory(package = "T0.WMBS",
                            logger = logging,
                            dbinterface = myThread.dbi)

    daoFactorySMNotify = DAOFactory(package = "T0.WMBS",
                                    logger = logging,
                                    dbinterface = dbInterfaceSMNotify)

    getFinishedStreamersDAO = daoFactory(classname = "SMNotification.GetFinishedStreamers")
    updateFileStatusDAO = daoFactorySMNotify(classname = "SMNotification.UpdateOfflineFileStatus")
    markStreamersFinishedDAO = daoFactory(classname = "SMNotification.MarkStreamersFinished")

    finishedStreamers = getFinishedStreamersDAO.execute(transaction = False)

    streamers = []
    bindUpdateFileStatus = []
    for (streamer_id, p5_id) in finishedStreamers:
        streamers.append(streamer_id)
        bindUpdateFileStatus.append( { 'P5_ID': p5_id } )

    if len(bindUpdateFileStatus) > 0:
        updateFileStatusDAO.execute(bindUpdateFileStatus, transaction = False)

    if len(streamers) > 0:
        markStreamersFinishedDAO.execute(streamers, transaction = False)

    return
